# Copyright 2023 by Garmin Ltd. or its subsidiaries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

import attr

from labgrid import target_factory, step
from labgrid.strategy import Strategy, StrategyError, never_retry
from labgrid.util import get_free_port


class Status(enum.Enum):
    unknown = 0
    off = 1
    shell = 2


@target_factory.reg_driver
@attr.s(eq=False)
class QEMUNetworkStrategy(Strategy):
    bindings = {
        "qemu": "QEMUDriver",
        "shell": "ShellDriver",
        "ssh": "SSHDriver",
    }

    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.__port_forward = None
        self.__remote_port = self.ssh.networkservice.port

    @step(result=True)
    def get_remote_address(self):
        return str(self.shell.get_ip_addresses()[0].ip)

    @step()
    def update_network_service(self):
        new_address = self.get_remote_address()
        networkservice = self.ssh.networkservice

        if networkservice.address != new_address:
            self.target.deactivate(self.ssh)

            if "user" in self.qemu.nic.split(","):
                if self.__port_forward is not None:
                    self.qemu.remove_port_forward(*self.__port_forward)

                local_port = get_free_port()
                local_address = "127.0.0.1"

                self.qemu.add_port_forward(
                    "tcp",
                    local_address,
                    local_port,
                    new_address,
                    self.__remote_port,
                )
                self.__port_forward = ("tcp", local_address, local_port)

                networkservice.address = local_address
                networkservice.port = local_port
            else:
                networkservice.address = new_address
                networkservice.port = self.__remote_port

    @never_retry
    @step(args=["state"])
    def transition(self, state, *, step):
        if not isinstance(state, Status):
            state = Status[state]

        if state == Status.unknown:
            raise StrategyError(f"can not transition to {state}")

        elif self.status == state:
            step.skip("nothing to do")
            return

        if state == Status.off:
            self.target.activate(self.qemu)
            self.qemu.off()

        elif state == Status.shell:
            self.target.activate(self.qemu)
            self.qemu.on()
            self.target.activate(self.shell)
            self.update_network_service()

        self.status = state



import attr

from ..factory import target_factory
from ..step import step
from ..util.managedfile import ManagedFile
from .common import Driver

from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class FlashScriptDriver(Driver):
    bindings = {
        "device": {
            "USBFlashableDevice",
            "NetworkUSBFlashableDevice",
        },
    }
    script = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    args = attr.ib(
        default=attr.Factory(list),
        validator=attr.validators.optional(attr.validators.instance_of(list)),
    )

    def on_activate(self):
        pass

    def on_deactivate(self):
        pass

    @Driver.check_active
    @step(args=["script"])
    def flash(self, script=None, args=None):
        """
        Transfers and remotely executes the script

        Args:
            script (str): optional, path to the script to write to bound Flashable Device
        """
        if script is None and self.script is not None:
            script = self.target.env.config.get_image_path(self.script)
        assert script, "flash requires a script"

        if args is None:
            args = self.args

        mf = ManagedFile(script, self.device)
        mf.sync_to_resource()

        cmd = [mf.get_remote_path()] + [a.format(device=self.device, file=mf) for a in args]

        self.logger.debug("Running command '%s'", " ".join(cmd))
        processwrapper.check_output(
            self.device.command_prefix + cmd, print_on_silent_log=True
        )



from time import sleep

from ..util import Timeout
from ..step import step
from .common import Driver
from .exception import ExecutionError


class CommandMixin:
    """
    CommandMixin implementing common functions for drivers which support the CommandProtocol
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @Driver.check_active
    @step(args=["cmd", "pattern"])
    def wait_for(self, cmd, pattern, timeout=30.0, sleepduration=1):
        """
        Wait until the pattern is detected in the output of cmd. Raises
        ExecutionError when the timeout expires.

        Args:
            cmd (str): command to run on the shell
            pattern (str): pattern as a string to look for in the output
            timeout (float): timeout for the pattern detection
            sleepduration (int): sleep time between the runs of the cmd
        """
        timeout = Timeout(timeout)
        while not any(pattern in s for s in self.run_check(cmd)) and not timeout.expired:
            sleep(sleepduration)
        if timeout.expired:
            raise ExecutionError("Wait timeout expired")

    @Driver.check_active
    @step(args=["cmd", "expected", "tries", "timeout", "sleepduration"])
    def poll_until_success(self, cmd, *, expected=0, tries=None, timeout=30.0, sleepduration=1):
        """
        Poll a command until a specific exit code is detected.
        Takes a timeout and the number of tries to run the cmd.
        The sleepduration argument sets the duration between runs of the cmd.

        Args:
            cmd (str): command to run on the shell
            expected (int): exitcode to detect
            tries (int): number of tries, can be None for infinite tries
            timeout (float): timeout for the exitcode detection
            sleepduration (int): sleep time between the runs of the cmd

        Returns:
            bool: whether the command finally executed sucessfully
        """
        timeout = Timeout(timeout)
        while not timeout.expired:
            _, _, exitcode = self.run(cmd, timeout=timeout.remaining)
            if exitcode == expected:
                return True
            sleep(sleepduration)
            if tries is not None:
                tries -= 1
                if tries < 1:
                    break
        return False

    def _run_check(self, cmd: str, *, timeout=30, codec: str = "utf-8", decodeerrors: str = "strict"):
        """
        Internal function which runs the specified command on the shell and
        returns the output if successful, raises ExecutionError otherwise.

        Args:
            cmd (str): command to run on the shell

        Returns:
            List[str]: stdout of the executed command
        """
        stdout, stderr, exitcode = self._run(cmd, timeout=timeout, codec=codec, decodeerrors=decodeerrors)
        if exitcode != 0:
            raise ExecutionError(cmd, stdout, stderr)
        return stdout

    @Driver.check_active
    @step(args=["cmd"], result=True)
    def run_check(self, cmd: str, *, timeout=30, codec="utf-8", decodeerrors="strict"):
        """
        External run_check function, only available if the driver is active.
        Runs the supplied command and returns the stdout, raises an
        ExecutionError otherwise.

        Args:
            cmd (str): command to run on the shell

        Returns:
            List[str]: stdout of the executed command
        """
        return self._run_check(cmd, timeout=timeout, codec=codec, decodeerrors=decodeerrors)



import pexpect


def test_autoinstall_help():
    with pexpect.spawn("python -m labgrid.autoinstall.main --help") as spawn:
        spawn.expect("usage")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


def test_autoinstall_error_missing_autoinstall(tmpdir):
    c = tmpdir.join("config.yaml")
    c.write("""
    targets:
        test: {}
    """)
    with pexpect.spawn(f"python -m labgrid.autoinstall.main {c}") as spawn:
        spawn.expect("no 'autoinstall' section found")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 1


def test_autoinstall_error_missing_handler(tmpdir):
    c = tmpdir.join("config.yaml")
    c.write("""
    autoinstall: |
        print("foo")
    """)
    with pexpect.spawn(f"python -m labgrid.autoinstall.main {c}") as spawn:
        spawn.expect("no 'handler' definition found")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 1


def test_autoinstall_no_targets(tmpdir):
    c = tmpdir.join("config.yaml")
    c.write("""
    autoinstall:
        handler: |
            print("handler-test-output")
    """)
    with pexpect.spawn(f"python -m labgrid.autoinstall.main {c}") as spawn:
        spawn.expect("no targets found")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 1


def test_autoinstall_simple(tmpdir):
    c = tmpdir.join("config.yaml")
    c.write("""
    targets:
        test1:
            resources: {}
            drivers: {}
    autoinstall:
        handler: |
            print("handler-test-output")
    """)
    with pexpect.spawn(f"python -m labgrid.autoinstall.main --once {c}") as spawn:
        spawn.expect("handler-test-output")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0



"""Interface for controlling relays of Shelly devices using the Gen 2+ API

  NetworkPowerPort:
      model: shelly_gen2
      host: 'http://192.168.0.42'
      index: 0

Will do a POST request to http://192.168.0.42/rpc to get the current
relay state or change the state.

Also, see the official Gen 2+ Device API documentation:
https://shelly-api-docs.shelly.cloud/gen2/General/RPCProtocol
"""

import requests


def power_set(host: str, port: int, index: int = 0, value: bool = True):
    assert not port
    payload = {"id": 1, "method": "Switch.Set", "params": {"id": index, "on": value}}
    r = requests.post(f"{host}/rpc", json=payload)
    r.raise_for_status()


def power_get(host: str, port: int, index: int = 0):
    assert not port
    payload = {"id": 1, "method": "Switch.GetStatus", "params": {"id": index}}
    r = requests.post(f"{host}/rpc", json=payload)
    r.raise_for_status()
    return r.json()["result"]["output"]



from labgrid.resource.modbusrtu import ModbusRTU
from labgrid.driver.modbusrtudriver import ModbusRTUDriver

import pytest

def test_resource_with_minimum_argument(target):
    dut = ModbusRTU(target, name=None, port="/dev/tty1", address=10)

    assert dut.port == "/dev/tty1"
    assert dut.address == 10
    assert dut.speed == 115200
    assert dut.timeout == 0.25


def test_resource_with_non_default_argument(target):
    dut = ModbusRTU(target, name=None, port="/dev/tty1", address=10,
                    speed=9600, timeout=0.5)

    assert dut.port == "/dev/tty1"
    assert dut.address == 10
    assert dut.speed == 9600
    assert dut.timeout == 0.5


def test_driver(target, mocker):
    pytest.importorskip("minimalmodbus")
    mocker.patch('serial.Serial')

    ModbusRTU(target, name=None, port="/dev/tty0", address=10)
    driver = ModbusRTUDriver(target, name=None)

    target.activate(driver)

    assert driver.instrument.serial.baudrate == 115200
    assert driver.instrument.serial.timeout == 0.25



"""
This driver implements a digital output driver for the robot electronics 8 relay
outputs board (ETH008).

Driver has been tested with:
* ETH008 - 8 relay outputs
"""

import socket
import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..step import step
from ..util.proxy import proxymanager
from .common import Driver
from .exception import ExecutionError

PORT = 17494  # TCP port for ETH008 (0x4456)


@target_factory.reg_driver
@attr.s(eq=False)
class Eth008DigitalOutputDriver(Driver, DigitalOutputProtocol):
    """Eth008DigitalOutputDriver - Driver to control individual relays on ETH008 as digital outputs"""
    bindings = {"output": {"Eth008DigitalOutput"}, }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._host = None
        self._port = None

    def on_activate(self):
        self._host, self._port = proxymanager.get_host_and_port(
            self.output, force_port=PORT
        )

    def _send_tcp_command(self, command_bytes):
        """Send a command over TCP and return the response."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(5.0)
            s.connect((self._host, self._port))
            s.sendall(command_bytes)
            response = s.recv(1)
            return response

    @Driver.check_active
    @step(args=["status"])
    def set(self, status):
        index = int(self.output.index)
        assert 1 <= index <= 8

        if self.output.invert:
            status = not status

        # Use TCP command: 0x20 for active (on), 0x21 for inactive (off)
        command = 0x20 if status else 0x21
        # Permanent mode (time = 0)
        command_bytes = bytes([command, index, 0])

        response = self._send_tcp_command(command_bytes)

        # Check response: 0 for success, 1 for failure
        if response == b'\x01':
            raise ExecutionError(f"failed to set port {index} to status {status}")
        elif response != b'\x00':
            raise ExecutionError(f"unexpected response from device: {response}")

    @Driver.check_active
    @step(result=["True"])
    def get(self):
        index = int(self.output.index)
        assert 1 <= index <= 8

        # Use TCP command: 0x24 to get all relay states
        # The response is 1 byte where each bit represents a relay state
        command_bytes = bytes([0x24])

        response = self._send_tcp_command(command_bytes)

        # Parse the response byte
        # Each bit represents a relay: bit 0 = relay 1, bit 1 = relay 2, etc.
        state_byte = response[0]
        # Get the bit corresponding to the relay index (1-8)
        relay_bit = (state_byte >> (index - 1)) & 0x01
        state = bool(relay_bit)

        if self.output.invert:
            state = not state

        return state



from .bootstrapprotocol import BootstrapProtocol
from .commandprotocol import CommandProtocol
from .consoleprotocol import ConsoleProtocol
from .linuxbootprotocol import LinuxBootProtocol
from .powerprotocol import PowerProtocol
from .filetransferprotocol import FileTransferProtocol
from .infoprotocol import InfoProtocol
from .digitaloutputprotocol import DigitalOutputProtocol
from .mmioprotocol import MMIOProtocol
from .filesystemprotocol import FileSystemProtocol
from .resetprotocol import ResetProtocol
from .videoprotocol import VideoProtocol



import logging
import time

from labgrid import Environment
from labgrid.logging import basicConfig, StepLogger

# enable info logging
basicConfig(level=logging.INFO)

# show labgrid steps on the console
StepLogger.start()

e = Environment("import-gpio.yaml")
t = e.get_target()

p = t.get_driver("DigitalOutputProtocol")
print(t.resources)
p.set(True)
print(p.get())
time.sleep(2)
p.set(False)
print(p.get())
time.sleep(2)
p.set(True)
print(p.get())



#!/usr/bin/env python3
"""Power-cycle a target until the /dev/nand0 device is missing."""

import sys
import logging

from labgrid import Environment
from labgrid.logging import basicConfig, StepLogger
from labgrid.strategy.bareboxstrategy import Status


# enable info logging
basicConfig(level=logging.INFO)

# log labgrid steps
StepLogger.start()


def run_once(target):
    s = target.get_driver("BareboxStrategy")
    s.status = Status.unknown  # force a power-cycle
    s.transition("barebox")
    cmd = target["CommandProtocol"]
    cmd.run_check("test -e /dev/nand0")
    target.deactivate(cmd)


env = Environment(sys.argv[1])
target = env.get_target("main")
while True:
    run_once(target)



import pytest


@pytest.fixture()
def signal_generator(target):
    return target.get_driver("PyVISADriver").get_session()


def test_with_signal_generator_example(signal_generator):
    signal_generator.write("*RST")

    # Setup channel 1
    signal_generator.write("C1:BSWV WVTP,SQUARE,HLEV,5,LLEV,0,DUTY,50")
    # Switch on channel 1
    signal_generator.write("C1:OUTP ON,LOAD,HZ,PLRT,NOR")



import re

import pytest

from labgrid.driver import ExecutionError


def test_rt_cyclictest_short(command):
    """Test a basic cyclictest run"""
    try:
        command.run_check("which cyclictest")
    except ExecutionError:
        pytest.skip("cyclictest missing")

    result = command.run_check("cyclictest -SN -D 5 -q")
    result = result[-1].strip()

    pattern = r"Min:\s+(?P<min>\w+)\s+Act:\s+\w+\s+Avg:\s+(?P<avg>\w+)\s+Max:\s+(?P<max>\w+)"
    min, avg, max = map(int, re.search(pattern, result).groups())
    assert min <= avg <= max
    assert avg < 1e6  # avg < 1 milliseconds
    assert max < 10e6  # max < 10 milliseconds


def test_rt_hackbench_short(command):
    """Test a basic hackbench run"""
    try:
        command.run_check("which hackbench")
    except ExecutionError:
        pytest.skip("hackbench missing")

    result = command.run_check("hackbench -f 10")
    result = result[-1].strip()

    pattern = r"Time:\s+(?P<time>\w+)"
    (time,) = map(int, re.search(pattern, result).groups())
    assert time < 20  # max 20 seconds



from importlib import import_module
from decimal import Decimal
import attr

from .common import Driver
from ..factory import target_factory
from ..exceptions import InvalidConfigError
from ..util.agentwrapper import AgentWrapper, b2s, s2b


@target_factory.reg_driver
@attr.s(eq=False)
class USBTMCDriver(Driver):
    bindings = {
        "tmc": {"USBTMC", "NetworkUSBTMC"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.wrapper = None
        self.backend = None

        assert self.tmc.path.startswith("/dev/usbtmc")
        self.index = int(self.tmc.path[11:], 10)
        assert self.tmc.path == "/dev/usbtmc" + str(self.index)

    def on_activate(self):
        assert self.wrapper is None
        self.wrapper = AgentWrapper(self.tmc.host)

        match = (self.tmc.vendor_id, self.tmc.model_id)
        if match == (0x0957, 0x1798):
            model = "keysight_dsox2000"
        elif match == (0x0699, 0x0368):
            model = "tektronix_tds2000"
        else:
            raise InvalidConfigError("Unkown USB TMC device {:04x}:{:04x}".format(*match))

        # TODO: allow backends to register models with other names
        self.backend = import_module(f".usbtmc.{model}", __package__)

    def on_deactivate(self):
        assert self.wrapper is not None
        self.wrapper.close()
        self.wrapper = None
        self.backend = None

    @Driver.check_active
    def command(self, cmd):
        assert isinstance(cmd, str)
        cmd = b2s(cmd.encode("ASCII") + b"\n")
        self.wrapper.usbtmc(self.index, cmd, read=False)

    @Driver.check_active
    def query(self, cmd, binary=False, raw=False):
        assert isinstance(cmd, str)
        cmd = b2s(cmd.encode("ASCII") + b"\n")
        res = s2b(self.wrapper.usbtmc(self.index, cmd, read=True))

        if raw:
            return res

        if binary:
            assert res[0:1] == b"#"
            digits = int(res[1:2], 10)
            count = int(res[2 : 2 + digits], 10)
            return res[2 + digits : 2 + digits + count]

        assert res[-1:] == b"\n"
        return res[:-1].decode("ASCII")

    @Driver.check_active
    def identify(self):
        return self.query("*IDN?")

    @Driver.check_active
    def get_channel_info(self, channel):
        return self.backend.get_channel_info(self, channel)

    @Driver.check_active
    def get_channel_values(self, channel):
        return self.backend.get_channel_values(self, channel)

    @Driver.check_active
    def get_screenshot(self):
        png = getattr(self.backend, "get_screenshot_png", None)
        tiff = getattr(self.backend, "get_screenshot_tiff", None)
        if png:
            return "png", png(self)  # pylint: disable=not-callable
        if tiff:
            return "tiff", tiff(self)  # pylint: disable=not-callable

        raise InvalidConfigError("get_screenshot_png/_tiff not implemented")

    @Driver.check_active
    def get_bool(self, cmd):
        return bool(int(self.query(cmd).strip()))

    @Driver.check_active
    def get_int(self, cmd):
        return int(self.query(cmd).strip())

    @Driver.check_active
    def get_decimal(self, cmd):
        return Decimal(self.query(cmd).strip())

    @Driver.check_active
    def get_str(self, cmd):
        return self.query(cmd)



import pytest
import socket

from labgrid.driver import Eth008DigitalOutputDriver
from labgrid.resource import Eth008DigitalOutput


@pytest.fixture(scope="function")
def mock_tcp_server(mocker):
    """Mock the ETH008 TCP server responses."""
    state = 0x00  # Initial state: all relays OFF

    # Track sent commands for verification
    sent_commands = []

    def mock_sendall(data):
        """Mock socket.sendall."""
        nonlocal state, sent_commands

        sent_commands.append(data)

        if len(data) == 1 and data[0] == 0x24:
            return
        elif len(data) == 3:
            command = data[0]
            relay_index = data[1]

            if command == 0x20:  # Digital Active (ON)
                state |= (1 << (relay_index - 1))
            elif command == 0x21:  # Digital Inactive (OFF)
                state &= ~(1 << (relay_index - 1))

    def mock_recv(bufsize):
        """Mock socket.recv to return appropriate responses."""
        nonlocal state, sent_commands

        last_sent = sent_commands[-1] if sent_commands else None

        if last_sent and len(last_sent) == 1 and last_sent[0] == 0x24:
            return bytes([state])
        else:
            return b'\x00'

    mock_socket = mocker.MagicMock()
    mock_socket.sendall.side_effect = mock_sendall
    mock_socket.recv.side_effect = mock_recv
    mock_socket.__enter__ = mocker.MagicMock(return_value=mock_socket)
    mock_socket.__exit__ = mocker.MagicMock(return_value=None)

    def mock_socket_factory(*args, **kwargs):
        return mock_socket

    mock_socket_class = mocker.patch("socket.socket")
    mock_socket_class.side_effect = mock_socket_factory
    mock_socket_class.return_value = mock_socket

    return mock_socket


def _make_eth008_driver(target, index, invert):
    """Helper function to create an ETH008 digital output driver with resource."""
    eth008_res = Eth008DigitalOutput(
        target,
        name=None,
        host='192.168.1.100',
        index=str(index),
        invert=invert
    )

    driver = Eth008DigitalOutputDriver(target, name=None)
    target.activate(driver)

    return driver


def test_eth008_instance(target, mocker):
    """Test that Eth008DigitalOutputDriver can be instantiated and activated."""
    mocker.patch("socket.socket")
    driver = _make_eth008_driver(target, 1, False)
    assert isinstance(driver, Eth008DigitalOutputDriver)


def test_eth008_set_true(target, mock_tcp_server):
    """Test setting ETH008 relay to True (asserted)."""
    driver = _make_eth008_driver(target, 1, False)

    driver.set(True)

    # Verify the correct command was sent (0x20 = Digital Active, 0x01 = relay 1, 0x00 = permanent)
    mock_tcp_server.sendall.assert_any_call(bytes([0x20, 0x01, 0x00]))


def test_eth008_set_false(target, mock_tcp_server):
    """Test setting ETH008 relay to False (de-asserted)."""
    driver = _make_eth008_driver(target, 2, False)

    driver.set(False)

    # Verify the correct command was sent (0x21 = Digital Inactive, 0x02 = relay 2, 0x00 = permanent)
    mock_tcp_server.sendall.assert_any_call(bytes([0x21, 0x02, 0x00]))


def test_eth008_get_true(target, mock_tcp_server):
    """Test getting ETH008 relay state when True."""
    driver = _make_eth008_driver(target, 1, False)

    # Set the relay to True first
    driver.set(True)

    # Get the state
    result = driver.get()

    assert result is True


def test_eth008_get_false(target, mock_tcp_server):
    """Test getting ETH008 relay state when False."""
    driver = _make_eth008_driver(target, 2, False)

    # Set the relay to False first
    driver.set(False)

    # Get the state
    result = driver.get()

    assert result is False


def test_eth008_invert_true(target, mock_tcp_server):
    """Test ETH008 with inversion enabled."""
    driver = _make_eth008_driver(target, 1, True)  # Invert the logic

    # When invert=True, setting True should actually set the relay to False
    driver.set(True)

    # Verify that 0x21 (Digital Inactive) was sent instead of 0x20 (Digital Active)
    mock_tcp_server.sendall.assert_any_call(bytes([0x21, 0x01, 0x00]))


def test_eth008_invert_get(target, mock_tcp_server):
    """Test getting ETH008 relay state with inversion enabled."""
    driver = _make_eth008_driver(target, 1, True)  # Invert the logic

    # Set the relay to True (which will actually set it to False due to inversion)
    driver.set(True)

    # When invert=True, get() should return the opposite of the actual state
    result = driver.get()

    # The actual relay state is False, but with inversion it should return True
    assert result is True


def test_eth008_invalid_index(target, mocker):
    """Test that invalid relay indices are handled correctly."""
    mocker.patch("socket.socket")

    driver = _make_eth008_driver(target, 9, False)  # Invalid index (must be 1-8)

    # This should raise an AssertionError due to invalid index
    with pytest.raises(AssertionError):
        driver.set(True)


def test_eth008_different_indices(target, mock_tcp_server):
    """Test that different relay indices work correctly."""
    # Test first, middle, and last relays
    # Create separate targets for each test to avoid resource conflicts
    from labgrid.target import Target
    for index in [1, 4, 8]:
        test_target = Target(name=f'Test-{index}', env=None)

        driver = _make_eth008_driver(test_target, index, False)

        # Set and get should work for all indices
        driver.set(True)
        result = driver.get()

        assert result is True



from collections import defaultdict

import attr


@attr.s(eq=False)
class TagSet:
    name = attr.ib(validator=attr.validators.instance_of(str))
    tags = attr.ib(validator=attr.validators.instance_of(set))


def schedule_step(places, filters):
    "Find the filters that can be directly allocated without overlap."
    interest = defaultdict(list)
    for f in filters:
        for place in places:
            if f.tags.issubset(place.tags):
                interest[place].append(f)

    if not interest:
        return {}

    limit = min(map(len, interest.values()))
    allocation = {}
    for interest_place, interest_filters in interest.items():
        if len(interest_filters) == limit:
            allocation[interest_filters.pop(0)] = interest_place

    return allocation


def schedule_overlaps(places, filters):
    "Iterate schedule_step until no more allocations are found."
    places = places[:]
    filters = filters[:]
    allocation = {}
    while True:
        new = schedule_step(places, filters)
        if not new:
            break
        for f, place in new.items():
            assert f not in allocation
            places.remove(place)
            filters.remove(f)
            allocation[f] = place
    return allocation


def schedule(places, filters):
    allocation = schedule_overlaps(places, filters)
    return {f.name: p.name for f, p in allocation.items()}



from labgrid.resource.httpvideostream import HTTPVideoStream
from labgrid.driver.httpvideodriver import HTTPVideoDriver


def test_ipvideo_create_http(target):
    r = HTTPVideoStream(target, name=None, url="http://localhost/")
    d = HTTPVideoDriver(target, name=None)
    assert isinstance(d, HTTPVideoDriver)

def test_ipvideo_create_https(target):
    r = HTTPVideoStream(target, name=None, url="https://localhost/")
    d = HTTPVideoDriver(target, name=None)
    assert isinstance(d, HTTPVideoDriver)

def test_ipvideo_create_with_port(target):
    r = HTTPVideoStream(target, name=None, url="http://localhost:8080/")
    d = HTTPVideoDriver(target, name=None)
    assert isinstance(d, HTTPVideoDriver)



def get_channel_info(driver, channel):
    info = {}
    info["BWLIMIT"] = driver.get_bool(f":CHANnel{channel:d}:BWLimit?")
    info["COUPLING"] = driver.get_str(f":CHANnel{channel:d}:COUPling?").strip()
    info["LABEL"] = driver.get_str(f":CHANnel{channel:d}:LABel?")
    info["PROBE"] = driver.get_decimal(f":CHANnel{channel:d}:PROBe?")
    info["PROBE:HEAD"] = driver.get_str(f":CHANnel{channel:d}:PROBe:HEAD?").strip()
    info["PROBE:ID"] = driver.get_str(f":CHANnel{channel:d}:PROBe:ID?").strip()
    info["PROBE:SKEW"] = driver.get_decimal(f":CHANnel{channel:d}:PROBe:SKEW?")
    info["PROBE:STYPE"] = driver.get_str(f":CHANnel{channel:d}:PROBe:STYPe?").strip()
    info["RANGE"] = driver.get_decimal(f":CHANnel{channel:d}:RANGe?")
    info["SCALE"] = driver.get_decimal(f":CHANnel{channel:d}:SCALe?")
    info["UNITS"] = driver.get_str(f":CHANnel{channel:d}:UNITS?").strip()
    return info


def get_channel_values(driver, channel):
    driver.command(f":MEASURE:SOURCE CHANNEL{channel:d}")
    info = {}
    info["DUTYCYCLE"] = driver.get_decimal(":MEASURE:DUTYcycle?")
    info["FREQUENCY"] = driver.get_decimal(":MEASURE:FREQuency?")
    info["PERIOD"] = driver.get_decimal(":MEASURE:PERiod?")
    info["VMAX"] = driver.get_decimal(":MEASURE:VMAX?")
    info["VMIN"] = driver.get_decimal(":MEASURE:VMIN?")
    info["VPP"] = driver.get_decimal(":MEASURE:VPP?")
    return info


def get_screenshot_png(driver):
    driver.command(":HARDcopy:INKSaver OFF")
    return driver.query(":DISPlay:DATA? PNG, COLor", binary=True)



from importlib import import_module
import attr

from ..factory import target_factory
from .common import Driver


@target_factory.reg_driver
@attr.s(eq=False)
class PyVISADriver(Driver):
    """The PyVISADriver provides a transparent layer to the PyVISA module allowing to get a pyvisa resource

    Args:
        bindings (dict): driver to use with PyVISA
    """

    bindings = {"pyvisa_resource": "PyVISADevice"}

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        _py_pyvisa_module = import_module("pyvisa")
        self._pyvisa_resource_manager = _py_pyvisa_module.ResourceManager(self.pyvisa_resource.backend)
        self.pyvisa_device = None

    def on_activate(self):
        device_identifier = f"{self.pyvisa_resource.type}::{self.pyvisa_resource.url}::INSTR"
        self.pyvisa_device = self._pyvisa_resource_manager.open_resource(device_identifier)

    def on_deactivate(self):
        self.pyvisa_device = None

    @Driver.check_active
    def get_session(self):
        return self.pyvisa_device



import logging
from signal import SIGTERM
import sys
import threading

import pytest
import pexpect

from labgrid import Target
from labgrid.driver import SerialDriver
from labgrid.resource import RawSerialPort, NetworkSerialPort
from labgrid.driver.fake import FakeConsoleDriver

psutil = pytest.importorskip("psutil")

@pytest.fixture(scope="session")
def curses_init():
    """ curses only reads the terminfo DB once on the first import, so make
    sure we prime it correctly."""
    try:
        import curses
        curses.setupterm("linux")
    except ModuleNotFoundError:
        logging.warning("curses module not found, not setting up a default terminal – tests may fail")


class Prefixer:
    def __init__(self, wrapped, prefix):
        self.__wrapped = wrapped
        self.__prefix = prefix.encode()+b": "
        self.__continuation = False

    def write(self, data):
        if data[-1:] == b'\n':
            continuation = False
            data = data[:-1]
        else:
            continuation = True
        data = data.replace(b'\n', b'\n'+self.__prefix)
        if not self.__continuation:
            data = self.__prefix+data
        if not continuation:
            data += b'\n'
        self.__continuation = continuation
        self.__wrapped.write(data)

    def __getattr__(self, name):
        return getattr(self.__wrapped, name)


class LabgridComponent:
    def __init__(self, cwd):
        self.cwd = str(cwd)
        self.spawn = None
        self.reader = None

    def stop(self):
        logging.info("stopping %s pid=%s", self.__class__.__name__, self.spawn.pid)

        # let coverage write its data:
        # https://coverage.readthedocs.io/en/latest/subprocess.html#process-termination
        self.spawn.kill(SIGTERM)
        if not self.spawn.closed:
            self.spawn.expect(pexpect.EOF)
            self.spawn.wait()
        assert not self.spawn.isalive()

        self.spawn = None
        self.stop_reader()

    @staticmethod
    def keep_reading(spawn):
        "The output from background processes must be read to avoid blocking them."
        while spawn.isalive():
            try:
                data = spawn.read_nonblocking(size=1024, timeout=0.1)
                if not data:
                    return
            except pexpect.TIMEOUT:
                continue
            except pexpect.EOF:
                return
            except OSError:
                return

    def start_reader(self):
        self.reader = threading.Thread(
            target=LabgridComponent.keep_reading,
            name=f'{self.__class__.__name__}-reader-{self.pid}',
            args=(self.spawn,), daemon=True)
        self.reader.start()

    def stop_reader(self):
        self.reader.join()

        self.reader = None

    def isalive(self):
        return self.spawn.isalive()

    @property
    def exitstatus(self):
        return self.spawn.exitstatus

    @property
    def pid(self):
        return self.spawn.pid

    def suspend_tree(self):
        main = psutil.Process(self.pid)
        main.suspend()
        for child in main.children(recursive=True):
            child.suspend()

    def resume_tree(self):
        main = psutil.Process(self.pid)
        main.resume()
        for child in main.children(recursive=True):
            child.resume()


class Exporter(LabgridComponent):
    def __init__(self, config, cwd):
        super().__init__(cwd)
        self.config = config

    def start(self):
        assert self.spawn is None
        assert self.reader is None

        self.spawn = pexpect.spawn(
            f'python -m labgrid.remote.exporter --name testhost {self.config}',
            logfile=Prefixer(sys.stdout.buffer, 'exporter'),
            cwd=self.cwd)
        try:
            self.spawn.expect('exporter name: testhost')
            self.spawn.expect('connected to coordinator')
        except Exception as e:
            raise Exception(f"exporter startup failed with {self.spawn.before}") from e

        self.start_reader()


class Coordinator(LabgridComponent):
    def start(self):
        assert self.spawn is None
        assert self.reader is None

        self.spawn = pexpect.spawn(
            'python -m labgrid.remote.coordinator',
            logfile=Prefixer(sys.stdout.buffer, 'coordinator'),
            cwd=self.cwd)
        try:
            self.spawn.expect('Coordinator ready')
        except Exception as e:
            raise Exception(f"coordinator startup failed with {self.spawn.before}") from e

        self.start_reader()


@pytest.fixture(scope='function')
def target():
    return Target('Test')

@pytest.fixture(scope='function')
def target_with_fakeconsole():
    t = Target('dummy')
    cp = FakeConsoleDriver(t, "console")
    return t

@pytest.fixture(scope='function')
def serial_port(target):
    return RawSerialPort(target, 'serial', '/dev/test')

@pytest.fixture(scope='function')
def serial_rfc2711_port(target):
    return NetworkSerialPort(target, 'rfc2711', host='localhost', port=8888)

@pytest.fixture(scope='function')
def serial_raw_port(target):
    return NetworkSerialPort(target, 'serialraw', host='localhost', port=8888, protocol="raw")


@pytest.fixture(scope='function')
def serial_driver(target, serial_port, mocker):
    m = mocker.patch('serial.Serial')
    s = SerialDriver(target, 'serial')
    target.activate(s)
    return s

@pytest.fixture(scope='function')
def serial_driver_no_name(target, serial_port, mocker):
    m = mocker.patch('serial.Serial')
    s = SerialDriver(target, None)
    target.activate(s)
    return s

@pytest.fixture(scope='function')
def coordinator(tmpdir):
    coordinator = Coordinator(tmpdir)
    coordinator.start()

    yield coordinator

    coordinator.stop()

@pytest.fixture(scope='function')
def exporter(tmpdir, coordinator):
    config = "exports.yaml"
    p = tmpdir.join(config)
    p.write(
        """
    Testport:
        NetworkSerialPort:
          host: 'localhost'
          port: 4000
    Broken:
        RawSerialPort:
          port: 'none'
    Many:
        NetworkSerialPort:
          host: 'localhost'
          port: 4000
        NetworkService:
          address: "192.168.0.1"
          username: "root"
    """
    )

    exporter = Exporter(config, tmpdir)
    exporter.start()

    yield exporter

    exporter.stop()

def pytest_addoption(parser):
    parser.addoption("--sigrok-usb", action="store_true",
                     help="Run sigrok usb tests with fx2lafw device (0925:3881)")
    parser.addoption("--local-sshmanager", action="store_true",
                     help="Run SSHManager tests against localhost")
    parser.addoption("--ssh-username", default=None,
                     help="SSH username to use for SSHDriver testing")

def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line("markers",
                            "sigrokusb: enable fx2lafw USB tests (0925:3881)")
    config.addinivalue_line("markers",
                            "localsshmanager: test SSHManager against Localhost")
    config.addinivalue_line("markers",
                            "sshusername: test SSHDriver against Localhost")
    config.addinivalue_line("markers",
                            "coordinator: test against local coordinator")

def pytest_runtest_setup(item):
    envmarker = item.get_closest_marker("sigrokusb")
    if envmarker is not None:
        if item.config.getoption("--sigrok-usb") is False:
            pytest.skip("sigrok usb tests not enabled (enable with --sigrok-usb)")
    envmarker = item.get_closest_marker("localsshmanager")
    if envmarker is not None:
        if item.config.getoption("--local-sshmanager") is False:
            pytest.skip("SSHManager tests against localhost not enabled (enable with --local-sshmanager)")
    envmarker = item.get_closest_marker("sshusername")
    if envmarker is not None:
        if item.config.getoption("--ssh-username") is None:
            pytest.skip("SSHDriver tests against localhost not enabled (enable with --ssh-username <username>)")



# pylint: disable=arguments-differ
import re

import attr

from ..factory import target_factory
from ..protocol import CommandProtocol, ConsoleProtocol, FileTransferProtocol, PowerProtocol
from .common import Driver
from .commandmixin import CommandMixin
from .consoleexpectmixin import ConsoleExpectMixin


@target_factory.reg_driver
@attr.s(eq=False)
class FakeConsoleDriver(ConsoleExpectMixin, Driver, ConsoleProtocol):
    txdelay = attr.ib(default=0.0, validator=attr.validators.instance_of(float))
    txchunk = attr.ib(default=1, validator=attr.validators.instance_of(int))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.rxq = []
        self.txq = []

    def _read(self, *_, **__):
        if self.rxq:
            return self.rxq.pop()
        return b''

    def _write(self, data, *_):
        self.txq.append(data)
        mo = re.match(rb'^echo "(\w+)""(\w+)"\n$', data)
        if mo:
            self.rxq.insert(0, b''.join(mo.groups())+b'\n')

    def open(self):
        pass

    def close(self):
        pass


@target_factory.reg_driver
@attr.s(eq=False)
class FakeCommandDriver(CommandMixin, Driver, CommandProtocol):
    @Driver.check_active
    def run(self, *args, timeout=None):
        pass

    @Driver.check_active
    def run_check(self, *args):
        pass

    @Driver.check_active
    def get_status(self):
        pass


@target_factory.reg_driver
@attr.s(eq=False)
class FakeFileTransferDriver(Driver, FileTransferProtocol):
    @Driver.check_active
    def get(self, *args):
        pass

    @Driver.check_active
    def put(self, *args):
        pass

@target_factory.reg_driver
@attr.s(eq=False)
class FakePowerDriver(Driver, PowerProtocol):
    @Driver.check_active
    def on(self, *args):
        pass

    @Driver.check_active
    def off(self, *args):
        pass

    @Driver.check_active
    def cycle(self, *args):
        pass



import subprocess
import re
import time

import attr

from ..factory import target_factory
from ..step import step
from .common import Driver
from .exception import ExecutionError
from ..util import Timeout
from ..util.helper import processwrapper
from ..util.managedfile import ManagedFile


@target_factory.reg_driver
@attr.s(eq=False)
class QuartusHPSDriver(Driver):
    bindings = {
        "interface": {"AlteraUSBBlaster", "NetworkAlteraUSBBlaster"},
    }

    image = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool("quartus_hps")
            self.jtag_tool = self.target.env.config.get_tool("jtagconfig")
        else:
            self.tool = "quartus_hps"
            self.jtag_tool = "jtagconfig"

    def _get_cable_number(self):
        """
        Returns the JTAG cable number with an intact JTAG chain for the USB path of the device.
        In case a matching JTAG cable is found, but its chain is broken, keep retrying for a
        while.
        """
        timeout = Timeout(10.0)
        while not timeout.expired:
            cmd = self.interface.command_prefix + [self.jtag_tool]
            jtagconfig_process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            stdout, _ = jtagconfig_process.communicate()

            regex = rf".*(\d+)\) .* \[{re.escape(self.interface.path)}]\n(.*)\n"
            jtag_mapping = re.search(regex, stdout.decode("utf-8"), re.MULTILINE)
            if jtag_mapping is None:
                raise ExecutionError(f"Could not get cable number for USB path {self.interface.path}")

            cable_number, first_chain = jtag_mapping.groups()
            try:
                jtag_id, _ = first_chain.split(sep="   ", maxsplit=1)
                int(jtag_id, 16)
            except ValueError:
                self.logger.warning("jtagconfig: %s", first_chain.strip())
                time.sleep(0.5)
                continue

            return int(cable_number)

        raise ExecutionError("Timeout while waiting for intact JTAG chain")

    @Driver.check_active
    @step(args=["filename", "address"])
    def flash(self, filename=None, address=0x0):
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        mf = ManagedFile(filename, self.interface)
        mf.sync_to_resource()

        assert isinstance(address, int)

        cable_number = self._get_cable_number()
        cmd = self.interface.command_prefix + [self.tool]
        cmd += [
            f"--cable={cable_number}",
            f"--addr=0x{address:X}",
            f"--operation=P {mf.get_remote_path()}",
        ]
        processwrapper.check_output(cmd)

    @Driver.check_active
    @step(args=["address", "size"])
    def erase(self, address=None, size=None):

        cable_number = self._get_cable_number()
        cmd = self.interface.command_prefix + [self.tool]
        cmd += [
            f"--cable={cable_number}",
            "--operation=E",
        ]
        if address:
            cmd += [f"--addr=0x{address:X}"]
        if size:
            cmd += [f"--size=0x{size:X}"]
        processwrapper.check_output(cmd)



import subprocess
import sys
from urllib.parse import urlsplit

import attr

from .common import Driver
from ..factory import target_factory
from ..util.proxy import proxymanager
from ..protocol import VideoProtocol


@target_factory.reg_driver
@attr.s(eq=False)
class HTTPVideoDriver(Driver, VideoProtocol):
    bindings = {
        "video": "HTTPVideoStream",
    }

    def get_qualities(self):
        return ("high", [("high", None)])

    @Driver.check_active
    def stream(self, quality_hint=None, controls=None):
        s = urlsplit(self.video.url)
        if s.port:
            default_port = s.port
        elif s.scheme == "http":
            default_port = 80
        elif s.scheme == "https":
            default_port = 443
        else:
            print(f"Unknown scheme: {s.scheme}", file=sys.stderr)
            return

        url = proxymanager.get_url(self.video.url, default_port=default_port)
        pipeline = [
            "gst-launch-1.0",
            "souphttpsrc",
            f"location={url}",
            "!",
            "decodebin",
            "!",
            "autovideoconvert",
            "!",
            "autovideosink",
            "sync=false",
        ]

        sub = subprocess.run(pipeline)
        return sub.returncode



def test_shell(target):
    ssh_driver = target.get_driver("SSHDriver")
    target.activate(ssh_driver)

    stdout, stderr, returncode = ssh_driver.run("uname -r")

    assert stdout
    print(f"Kernel {stdout}")



"""
This driver implements a power port for the robot electronics 8 relay
outputs board.

Driver has been tested with:
* ETH008 - 8 relay outputs
"""

import requests
from ..exception import ExecutionError

PORT = 80


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 8
    # access the web interface...
    value_str = "A" if value else "I"
    response = requests.get(f"http://{host}:{port}/io.cgi?DO{value_str}{index}")
    response.raise_for_status()

    # Check, that the port is in the desired state
    state = get_state(response, index)
    if state != value:
        raise ExecutionError(f"failed to set port {index} to status {value}")


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 8
    # get the contents of the main page
    response = requests.get(f"http://{host}:{port}/io.cgi?relay")

    response.raise_for_status()
    state = get_state(response, index)
    return state


def get_state(request, index):
    value = request.text.split()[1][index - 1]
    return bool(int(value))



# pylint: disable=unused-argument
"""The ShellDriver provides the CommandProtocol, ConsoleProtocol and
 InfoProtocol on top of a SerialPort."""
import io
import re
import shlex
import ipaddress

import attr
from pexpect import TIMEOUT
import xmodem

from ..factory import target_factory
from ..protocol import CommandProtocol, ConsoleProtocol, FileTransferProtocol
from ..step import step
from ..util import gen_marker, Timeout, re_vt100
from .commandmixin import CommandMixin
from .common import Driver
from .exception import ExecutionError


@target_factory.reg_driver
@attr.s(eq=False)
class ShellDriver(CommandMixin, Driver, CommandProtocol, FileTransferProtocol):
    """ShellDriver - Driver to execute commands on the shell
    ShellDriver binds on top of a ConsoleProtocol.

    On activation, the ShellDriver will look for the login prompt on the console,
    and login to provide shell access.

    Args:
        prompt (regex): the shell prompt to detect
        login_prompt (regex): the login prompt to detect
        username (str): username to login with
        password (str): password to login with
        keyfile (str): keyfile to bind mount over users authorized keys
        login_timeout (int): optional, timeout for login prompt detection
        console_ready (regex): optional, pattern used by the kernel to inform the user that a
            console can be activated by pressing enter.
        await_login_timeout (int): optional, time in seconds of silence that needs to pass before
            sending a newline to device.
        post_login_settle_time (int): optional, seconds of silence after logging in
            before check for a prompt. Useful when the console is interleaved with boot
            output which may interrupt prompt detection.
    """
    bindings = {"console": ConsoleProtocol, }
    prompt = attr.ib(validator=attr.validators.instance_of(str))
    login_prompt = attr.ib(validator=attr.validators.instance_of(str))
    username = attr.ib(validator=attr.validators.instance_of(str))
    password = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    keyfile = attr.ib(default="", validator=attr.validators.instance_of(str))
    login_timeout = attr.ib(default=60, validator=attr.validators.instance_of(int))
    console_ready = attr.ib(default="", validator=attr.validators.instance_of(str))
    await_login_timeout = attr.ib(default=2, validator=attr.validators.instance_of(int))
    post_login_settle_time = attr.ib(default=0, validator=attr.validators.instance_of(int))


    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._status = 0

        self._xmodem_cached_rx_cmd = ""
        self._xmodem_cached_sx_cmd = ""

    def on_activate(self):
        if self._status == 0:
            self._await_login()
            self._inject_run()

        if self.keyfile:
            keyfile_path = self.keyfile
            if self.target.env:
                keyfile_path = self.target.env.config.resolve_path(self.keyfile)

            self._put_ssh_key(keyfile_path)

    def on_deactivate(self):
        self._status = 0

    def _run(self, cmd, *, timeout=30.0, codec="utf-8", decodeerrors="strict"):
        """
        Runs the specified cmd on the shell and returns the output.

        Arguments:
        cmd - cmd to run on the shell
        """
        # FIXME: Handle pexpect Timeout
        self._check_prompt()
        marker = gen_marker()
        # hide marker from expect
        cmp_command = f'''MARKER='{marker[:4]}''{marker[4:]}' run {shlex.quote(cmd)}'''
        self.console.sendline(cmp_command)
        _, _, match, _ = self.console.expect(
            rf'{marker}(.*){marker}\s+(\d+)\s+{self.prompt}',
            timeout=timeout
        )
        # Remove VT100 Codes, split by newline and remove surrounding newline
        data = re_vt100.sub('', match.group(1).decode(codec, decodeerrors)).split('\r\n')
        if data and not data[-1]:
            del data[-1]
        self.logger.debug("Received Data: %s", data)
        # Get exit code
        exitcode = int(match.group(2))
        return (data, [], exitcode)

    @Driver.check_active
    @step(args=['cmd'], result=True)
    def run(self, cmd, timeout=30.0, codec="utf-8", decodeerrors="strict"):
        return self._run(cmd, timeout=timeout, codec=codec, decodeerrors=decodeerrors)

    @step()
    def _await_login(self):
        """Awaits the login prompt and logs the user in"""

        timeout = Timeout(float(self.login_timeout))

        expectations = [self.prompt, self.login_prompt, "Password: ", TIMEOUT]
        if self.console_ready != "":
            expectations.append(self.console_ready)

        # We call console.expect with a short timeout here to detect if the
        # console is idle, which results in a timeout without any changes to
        # the before property. So we store the last before value we've seen.
        # Because pexpect keeps any read data in it's buffer when a timeout
        # occours, we can't lose any data this way.
        last_before = b''
        did_login = False
        did_silence_kernel = False

        while True:
            index, before, _, _ = self.console.expect(
                expectations,
                timeout=self.await_login_timeout
            )

            if index == 0:
                if did_login and not did_silence_kernel:
                    # Silence the kernel and wait for another prompt
                    self.console.sendline("dmesg -n 1")
                    did_silence_kernel = True
                else:
                    # we got a prompt. no need for any further action to
                    # activate this driver.
                    self._status = 1
                    break

            elif index == 1:
                # we need to login
                self.console.sendline(self.username)
                did_login = True

            elif index == 2:
                if self.password is not None:
                    self.console.sendline(self.password)
                else:
                    raise Exception("Password entry needed but no password set")

            elif index == 3:
                # expect hit a timeout while waiting for a match
                if before == last_before:
                    # we did not receive anything during
                    # self.await_login_timeout.
                    # let's assume the target is idle and we can safely issue a
                    # newline to check the state
                    self.console.sendline("")

            elif index == 4:
                # we have just activated a console here
                # lets start over again and see if login or prompt will appear
                # now.
                self.console.sendline("")
                did_login = True

            last_before = before

            if timeout.expired:
                raise TIMEOUT(f"Timeout of {self.login_timeout} seconds exceeded during waiting for login")  # pylint: disable=line-too-long

        if did_login:
            if self.post_login_settle_time > 0:
                self.console.settle(self.post_login_settle_time, timeout=timeout.remaining)
            self._check_prompt()

    @step()
    def get_status(self):
        """Returns the status of the shell-driver.
        0 means not connected/found, 1 means shell
        """
        return self._status

    def _check_prompt(self):
        """
        Internal function to check if we have a valid prompt
        """
        marker = gen_marker()
        # hide marker from expect
        self.console.sendline(f"echo '{marker[:4]}''{marker[4:]}'")
        try:
            self.console.expect(
                rf"{marker}\s+{self.prompt}",
                timeout=30
            )
            self._status = 1
        except TIMEOUT:
            self._status = 0
            raise

    def _inject_run(self):
        self.console.sendline(
            '''run() { echo -n "$MARKER"; sh -c "$@"; echo "$MARKER $?"; }'''
        )
        self.console.expect(self.prompt)

    def _write_key(self, keyline, dest):
        for i in range(0, len(keyline), 100):
            part = keyline[i:i+100]
            self._run_check(f'echo -n "{part}" >> {dest}')
        self._run_check(f'echo "" >> {dest}')

    @step(args=['keyfile_path'])
    def _put_ssh_key(self, keyfile_path):
        """Upload an SSH Key to a target"""
        regex = re.compile(
            r"""ssh-(rsa|ed25519)
            \s+(?P<key>[a-zA-Z0-9/+=]+) # Match Keystring
            \s+(?P<comment>.*) # Match comment""", re.X
        )
        with open(keyfile_path) as keyfile:
            keyline = keyfile.readline()
            self.logger.debug("Read Keyline: %s", keyline)
            match = regex.match(keyline)
            if match:
                new_key = match.groupdict()
            else:
                raise IOError(
                    f"Could not parse SSH-Key from file: {keyfile}"
                )
        self.logger.debug("Read Key: %s", new_key)
        auth_keys, _, read_keys = self._run("cat ~/.ssh/authorized_keys")
        self.logger.debug("Exitcode trying to read keys: %s, keys: %s", read_keys, auth_keys)
        result = []
        _, _, test_write = self._run("touch ~/.test")
        if read_keys == 0:
            for line in auth_keys:
                match = regex.match(line)
                if match:
                    match = match.groupdict()
                    self.logger.debug("Match dict: %s", match)
                    result.append(match)
            self.logger.debug("Complete result: %s", result)
            for key in result:
                self.logger.debug(
                    "Key, newkey: %s,%s", key['key'], new_key['key']
                )
                if key['key'] == new_key['key']:
                    self.logger.debug("Key already on target")
                    return

        if test_write == 0 and read_keys == 0:
            self.logger.debug("Key not on target and writeable, concatenating...")
            self._write_key(keyline, "~/.ssh/authorized_keys")
            self._run_check("rm ~/.test")
            return

        if test_write == 0:
            self.logger.debug("Key not on target, testing for .ssh directory")
            _, _, ssh_dir = self._run("[ -d ~/.ssh/ ]")
            if ssh_dir != 0:
                self.logger.debug("~/.ssh did not exist, creating")
                self._run("mkdir ~/.ssh/")
            self._run_check("chmod 700 ~/.ssh/")
            self.logger.debug("Creating ~/.ssh/authorized_keys")
            self._run_check("touch ~/.ssh/authorized_keys")
            self._write_key(keyline, "~/.ssh/authorized_keys")
            self._run_check("rm ~/.test")
            return

        self.logger.debug("Key not on target and not writeable, using bind mount...")
        self._run_check('mkdir -p -m 700 /tmp/labgrid-ssh/')
        self._run("cp -aL ~/.ssh/* /tmp/labgrid-ssh/")
        self._write_key(keyline, "/tmp/labgrid-ssh/authorized_keys")
        self._run_check('chmod 600 /tmp/labgrid-ssh/authorized_keys')
        out, err, exitcode = self._run('mount --bind /tmp/labgrid-ssh/ ~/.ssh/')
        if exitcode != 0:
            self.logger.warning("Could not bind mount ~/.ssh directory: %s %s", out, err)

    @Driver.check_active
    def put_ssh_key(self, keyfile_path):
        self._put_ssh_key(keyfile_path)

    def _xmodem_getc(self, size, timeout=10):
        """ called by the xmodem.XMODEM instance to read protocol data from the console """
        try:
            # use the underlying expect mechanism, which may have already accidentally read
            # something of the XMODEM protocol data into its internal buffers:
            xpct = self.console.expect(r'.{%d}' % size, timeout=timeout)
            s = xpct[2].group()
            self.logger.debug('XMODEM GETC(%d): read %r', size, s)
            return s
        except TIMEOUT:
            self.logger.debug('XMODEM GETC(%s): TIMEOUT after %d seconds', size, timeout)
            return None

    def _xmodem_putc(self, data, timeout=1):
        """ called by the xmodem.XMODEM instance to write protocol data to the console """
        # Note: we ignore the timeout because we cannot pass it through.
        self.logger.debug('XMODEM PUTC: %r', data)
        self.console.write(data)
        return len(data)

    def _start_xmodem_transfer(self, cmd):
        """
        Start transfer command and synchronize until start of XMODEM stream.

        We don't use _run() here because it expects a prompt, but we want to
        read from the console directly into our XMODEM instance instead.
        """

        marker = gen_marker()
        marked_cmd = f"echo -n '{marker[:4]}''{marker[4:]}'; {cmd}"
        self.console.sendline(marked_cmd)
        self.console.expect(marker, timeout=30)

    def _get_xmodem_rx_cmd(self, filename):
        """ Detect which XMODEM receive command can be used on the target, and cache the result. """
        if not self._xmodem_cached_rx_cmd:
            if self._run('which lrz')[2] == 0:
                # redirect stderr to prevent lrz from printing "ready to receive
                # $file", which will confuse the XMODEM instance
                self._xmodem_cached_rx_cmd = "lrz -X -y -c -b '{filename}' 2>/dev/null"
            elif self._run('which rz')[2] == 0:
                # renamed binaries packaged by some distros
                self._xmodem_cached_rx_cmd = "rz -X -y -c -b '{filename}' 2>/dev/null"
            elif self._run('which rx')[2] == 0:
                # busybox rx
                # lrz may provide rx so redirect stderr for the same reason as above
                self._xmodem_cached_rx_cmd = "rx '{filename}' 2>/dev/null"
            else:
                raise ExecutionError('No XMODEM receiver (lrz, rz, rx) available on target')

        # use the cached string template to make the full command with parameters
        return self._xmodem_cached_rx_cmd.format(filename=filename)

    def _get_xmodem_sx_cmd(self, filename):
        """ Detect which XMODEM send command can be used on the target, and cache the result. """
        if not self._xmodem_cached_sx_cmd:
            if self._run('which lsz')[2] == 0:
                # redirect stderr to prevent lsz from printing "Give XMODEM receive
                # cmd now", which will confuse the XMODEM instance
                self._xmodem_cached_sx_cmd = "lsz -b -X -m 1200 -M 10 '{filename}' 2>/dev/null"
            elif self._run('which sz')[2] == 0:
                # renamed binaries packaged by some distros
                self._xmodem_cached_sx_cmd = "sz -b -X -m 1200 -M 10 '{filename}' 2>/dev/null"
            else:
                raise ExecutionError('No XMODEM sender (lsz, sz) available on target')

        # use the cached string template to make the full command with parameters
        return self._xmodem_cached_sx_cmd.format(filename=filename)

    @step(title='put_bytes', args=['remotefile'])
    def _put_bytes(self, buf: bytes, remotefile: str):
        # OK, a little explanation on what we're doing here:
        # XMODEM is a fairly simple, but also a fairly historic protocol. For example, all packets
        # carry exactly 128 bytes of payload, and if the file being sent is not a multiple of 128
        # bytes, the last packet will be padded by CPM's EOF, which is 0x1a. There is no file size
        # or anything in the protocol itself, so we'll have to take care of that and truncate the
        # file ourselves.

        def _target_cleanup(tmpfile):
            self._run(f"rm -f '{tmpfile}'")

        stream = io.BytesIO(buf)

        # We first write to a temp file, which we'll `dd` onto the destination file later
        tmpfile = self._run_check('mktemp')
        if not tmpfile:
            raise ExecutionError('Could not make temporary file on target')
        tmpfile = tmpfile[0]

        try:
            rx_cmd = self._get_xmodem_rx_cmd(tmpfile)
            self.logger.debug('XMODEM receive command on target: %s', rx_cmd)
        except ExecutionError:
            _target_cleanup(tmpfile)
            raise

        self._start_xmodem_transfer(rx_cmd)

        modem = xmodem.XMODEM(self._xmodem_getc, self._xmodem_putc)
        ret = modem.send(stream)
        self.logger.debug('xmodem.send() returned %r', ret)

        self.console.expect(self.prompt, timeout=30)

        # truncate the file to get rid of CPMEOF padding
        dd_cmd = f"dd if='{tmpfile}' of='{remotefile}' bs=1 count={len(buf)}"
        self.logger.debug('dd command: %s', dd_cmd)
        out, _, ret = self._run(dd_cmd)

        _target_cleanup(tmpfile)
        if ret != 0:
            raise ExecutionError(f'Could not truncate destination file: dd returned {ret}: {out}')

    @Driver.check_active
    def put_bytes(self, buf: bytes, remotefile: str):
        """ Upload a file to the target.
        Will silently overwrite the remote file if it already exists.

        Args:
            buf (bytes): file contents
            remotefile (str): destination filename on the target

        Raises:
            ExecutionError: if something went wrong
        """
        return self._put_bytes(buf, remotefile)

    @step(title='put', args=['localfile', 'remotefile'])
    def _put(self, localfile: str, remotefile: str):
        with open(localfile, 'rb') as fh:
            buf = fh.read(None)
            self._put_bytes(buf, remotefile)

    @Driver.check_active
    def put(self, localfile: str, remotefile: str):
        """ Upload a file to the target.
        Will silently overwrite the remote file if it already exists.

        Args:
            localfile (str): source filename on the local machine
            remotefile (str): destination filename on the target

        Raises:
            IOError: if the provided localfile could not be found
            ExecutionError: if something else went wrong
        """
        self._put(localfile, remotefile)

    @step(title='get_bytes', args=['remotefile'])
    def _get_bytes(self, remotefile: str):
        buf = io.BytesIO()

        cmd = self._get_xmodem_sx_cmd(remotefile)
        self.logger.info('XMODEM send command on target: %s', cmd)

        # get file size to remove XMODEM's CPMEOF padding at the end of the last packet
        out, _, ret = self._run(f"stat '{remotefile}'")
        match = re.search(r'Size:\s+(?P<size>\d+)', '\n'.join(out))
        if ret != 0 or not match or not match.group("size"):
            raise ExecutionError(f"Could not stat '{remotefile}' on target")

        file_size = int(match.group('size'))
        self.logger.debug('file size on target is %d', file_size)

        self._start_xmodem_transfer(cmd)

        modem = xmodem.XMODEM(self._xmodem_getc, self._xmodem_putc)
        recvd_size = modem.recv(buf)
        self.logger.debug('xmodem.recv() returned %r', recvd_size)

        # remove CPMEOF (0x1a) padding
        if recvd_size < file_size:
            raise ExecutionError(f'Only received {recvd_size} bytes of {file_size} expected')

        self.logger.debug('received %d bytes of payload', file_size)
        buf.truncate(file_size)

        self.console.expect(self.prompt, timeout=30)

        # return everything as bytes
        buf.seek(0)
        return buf.read()

    @Driver.check_active
    def get_bytes(self, remotefile: str):
        """ Download a file from the target.

        Args:
            remotefile (str): source filename on the target

        Returns:
            (bytes) file contents

        Raises:
            ExecutionError: if something went wrong
        """
        return self._get_bytes(remotefile)

    @step(title='get', args=['remotefile', 'localfile'])
    def _get(self, remotefile: str, localfile: str):
        with open(localfile, 'wb') as fh:
            buf = self._get_bytes(remotefile)
            fh.write(buf)

    @Driver.check_active
    def get(self, remotefile: str, localfile: str):
        """ Download a file from the target.
        Will silently overwrite the local file if it already exists.

        Args:
            remotefile (str): source filename on the target
            localfile (str): destination filename on the local machine (can be relative)

        Raises:
            IOError: if localfile could not be written
            ExecutionError: if something went wrong
        """
        self._get(remotefile, localfile)

    @step(title='run_script', args=['data', 'timeout'])
    def _run_script(self, data: bytes, timeout: int = 60):
        hardcoded_remote_file = '/tmp/labgrid-run-script'
        self._put_bytes(data, hardcoded_remote_file)
        self._run_check(f"chmod +x '{hardcoded_remote_file}'")
        return self._run(hardcoded_remote_file, timeout=timeout)

    @Driver.check_active
    def run_script(self, data: bytes, timeout: int = 60):
        """ Upload a script to the target and run it.

        Args:
            data (bytes): script data
            timeout (int): timeout for the script to finish execution

        Returns:
            Tuple of (stdout: str, stderr: str, return_value: int)

        Raises:
            ExecutionError: if something went wrong
        """
        return self._run_script(data, timeout)

    @step(title='run_script_file', args=['scriptfile', 'timeout', 'args'])
    def _run_script_file(self, scriptfile: str, *args, timeout: int = 60):
        hardcoded_remote_file = '/tmp/labgrid-run-script'
        self._put(scriptfile, hardcoded_remote_file)
        self._run_check(f"chmod +x '{hardcoded_remote_file}'")

        shargs = [shlex.quote(a) for a in args]
        cmd = f"{hardcoded_remote_file} {' '.join(shargs)}"
        return self._run(cmd, timeout=timeout)

    @Driver.check_active
    def run_script_file(self, scriptfile: str, *args, timeout: int = 60):
        """ Upload a script file to the target and run it.

        Args:
            scriptfile (str): source file on the local file system to upload to the target
            *args: (list of str): any arguments for the script as positional arguments
            timeout (int): timeout for the script to finish execution

        Returns:
            Tuple of (stdout: str, stderr: str, return_value: int)

        Raises:
            ExecutionError: if something went wrong
            IOError: if the provided localfile could not be found
        """
        return self._run_script_file(scriptfile, *args, timeout=timeout)

    @Driver.check_active
    def get_default_interface_device_name(self, version=4):
        """ Retrieve the default route's interface device name.

        Args:
            version (int): IP version

        Returns:
            Name of the device of the default route

        Raises:
            ExecutionError: if no or multiple routes are set up
        """
        assert version in (4, 6)

        regex = r"""default\s+via # leading strings
                \s+\S+ # IP address
                \s+dev\s+([\w\.-]+) # interface"""

        default_route = self._run_check(f"ip -{version} route list default")
        matches = re.findall(regex, "\n".join(default_route), re.X)

        if not matches:
            raise ExecutionError(f"No IPv{version} default route found")
        if len(matches) > 1:
            raise ExecutionError(f"Multiple IPv{version} default routes found")

        return matches[0]

    @Driver.check_active
    def get_ip_addresses(self, device=None):
        """ Retrieves IP addresses for given interface name.

        Note that although the return type is named IPv4Interface/IPv6Interface, it contains an IP
        address with the corresponding network prefix.

        Args:
            device (str): Name of the interface to query, defaults to default route's device name.

        Returns:
            List of IPv4Interface or IPv6Interface objects
        """
        if device is None:
            device = self.get_default_interface_device_name()

        regex = r"""\d+: # leading number
                \s+[\w\.-]+ # interface name
                \s+inet6?\s+(\S+) # IP address, prefix
                .*global # global scope, not host scope"""

        ip_show = self._run_check(f"ip -o addr show dev {device}")
        matches = re.findall(regex, "\n".join(ip_show), re.X)

        return list(map(ipaddress.ip_interface, matches))



import pytest
import subprocess
import tempfile
import attr
import os
from pathlib import Path
from labgrid.driver.flashscriptdriver import FlashScriptDriver
from labgrid.resource.common import ManagedResource
from labgrid import target_factory


@target_factory.reg_resource
@attr.s(eq=False)
class MockFlashableDevice(ManagedResource):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.avail = True

    @property
    def prop(self):
        return "test"


@pytest.fixture(scope="function")
def resource(target):
    r = MockFlashableDevice(target, name=None)
    assert isinstance(r, MockFlashableDevice)
    target.activate(r)
    return r


@pytest.fixture(scope="function")
def driver(target, mocker, resource):
    mocker.patch.dict(
        FlashScriptDriver.bindings,
        {
            "device": {
                "MockFlashableDevice",
            },
        },
    )

    d = FlashScriptDriver(target, name=None)
    assert isinstance(d, FlashScriptDriver)
    target.activate(d)
    return d


def capture_argument_expansion(d, var):
    with tempfile.NamedTemporaryFile() as f:
        d.flash("/bin/sh", args=["-c", "echo -n {%s} > %s" % (var, f.name)])
        return f.read().decode("utf-8")


@pytest.mark.skipif(not Path("/bin/true").exists(), reason="true not available")
def test_script_success(target, driver):
    driver.flash("/bin/true")


@pytest.mark.skipif(not Path("/bin/false").exists(), reason="false not available")
def test_script_failure(target, driver):
    with pytest.raises(subprocess.CalledProcessError):
        driver.flash("/bin/false")


def test_argument_device_expansion(target, resource, driver):
    value = capture_argument_expansion(driver, "device.prop")
    assert value == resource.prop


def test_argument_file_expansion(target, driver):
    value = capture_argument_expansion(driver, "file.local_path")
    assert os.path.samefile(value, "/bin/sh")



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class XenaManager(Resource):
    """Hostname/IP identifying the management address of the xena tester"""

    hostname = attr.ib(validator=attr.validators.instance_of(str))



import abc


class PowerProtocol(abc.ABC):
    @abc.abstractmethod
    def on(self):
        raise NotImplementedError

    @abc.abstractmethod
    def off(self):
        raise NotImplementedError

    @abc.abstractmethod
    def cycle(self):
        raise NotImplementedError



import attr

from ..factory import target_factory
from .common import Resource, NetworkResource
from .base import SerialPort


@target_factory.reg_resource
@attr.s(eq=False)
class RawSerialPort(SerialPort, Resource):
    """RawSerialPort describes a serialport which is available on the local computer."""

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.port is None:
            raise ValueError("RawSerialPort must be configured with a port")


# This does not derive from SerialPort because it is not directly accessible
@target_factory.reg_resource
@attr.s(eq=False)
class NetworkSerialPort(NetworkResource):
    """A NetworkSerialPort is a remotely accessible serialport, usually
    accessed via rfc2217 or tcp raw.

    Args:
        port (str): socket port to connect to
        speed (int): speed of the port e.g. 9800
        protocol (str): connection protocol: "raw" or "rfc2217"
    """

    port = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(int)))
    speed = attr.ib(default=115200, validator=attr.validators.instance_of(int))
    protocol = attr.ib(default="rfc2217", validator=attr.validators.instance_of(str))



import logging
import subprocess
import attr

from ..binding import BindingError, BindingMixin
from .exception import ExecutionError


@attr.s(eq=False)
class Driver(BindingMixin):
    """
    Represents a driver which is used externally or by other drivers. It
    implements functionality based on directly accessing the Resource or by
    building on top of other Drivers.

    Life cycle:
    - create
    - bind (n times)
    - activate
    - usage
    - deactivate
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target is None:
            raise BindingError("Drivers can only be created on a valid target")

        logger_name = f"{self.__class__.__name__}({self.target.name})"
        if self.name:
            logger_name += f":{self.name}"
        self.logger = logging.getLogger(logger_name)

    def get_priority(self, protocol):
        """Retrieve the priority for a given protocol

        Arguments:
        protocol - protocol to search for in the MRO

        Returns:
            Int: value of the priority if it is found, 0 otherwise.
        """
        for cls in self.__class__.__mro__:
            prios = getattr(cls, 'priorities', {})
            # we found a matching parent priorities attribute with the matching protocol
            if prios and protocol in prios:
                return prios.get(protocol)
            # If we find the parent protocol, set the priority to 0
            if cls.__name__ == protocol.__name__:
                return 0

        return 0

    def get_export_name(self):
        """Get the name to be used for exported variables.

        Falls back to the class name if the driver has no name.
        """
        if self.name:
            return self.name
        return self.__class__.__name__

    def get_export_vars(self):
        """Get a dictionary of variables to be exported."""
        return {}

    @property
    def skip_deactivate_on_export(self):
        """Drivers are deactivated on export by default.

        If the driver can handle external accesses even while active, it can
        return True here.
        """
        return False

    def get_bound_resources(self):
        """Return the bound resources for a driver

        This recursively calls all suppliers and combines the sets of returned resources.
        """
        res = set()
        for supplier in self.suppliers:
            res |= supplier.get_bound_resources()
        return res

def check_file(filename, *, command_prefix=[]):
    if subprocess.call(command_prefix + ['test', '-r', filename]) != 0:
        raise ExecutionError(f"File {filename} is not readable")



import subprocess

import attr

from .common import Driver
from ..factory import target_factory
from ..step import step
from .exception import ExecutionError
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class USBSDMuxDriver(Driver):
    """The USBSDMuxDriver uses the usbsdmux tool
    (https://github.com/pengutronix/usbsdmux) to control the USB-SD-Mux
    hardware

    Args:
        bindings (dict): driver to use with usbsdmux
    """

    bindings = {
        "mux": {"USBSDMuxDevice", "NetworkUSBSDMuxDevice"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool("usbsdmux")
        else:
            self.tool = "usbsdmux"

    @Driver.check_active
    @step(title="sdmux_set", args=["mode"])
    def set_mode(self, mode):
        if not mode.lower() in ["dut", "host", "off", "client"]:
            raise ExecutionError(f"Setting mode '{mode}' not supported by USBSDMuxDriver")
        cmd = self.mux.command_prefix + [self.tool, self.mux.control_path, mode.lower()]
        processwrapper.check_output(cmd)

    @Driver.check_active
    @step(title="sdmux_get")
    def get_mode(self):
        cmd = self.mux.command_prefix + [self.tool, self.mux.control_path, "get"]
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
        return proc.stdout.strip().decode()



import pytest
import pexpect


@pytest.fixture
def env_feature_config(tmpdir):
    def _env_feature_config(features):
        conf = tmpdir.join("config.yaml")
        conf.write(
            """
    targets:
      test1:
        features:
"""
        )
        for feature in features:
            conf.write(f"          - {feature}\n", "a")
        if not features:
            conf.write("          {}\n", "a")

        return conf

    yield _env_feature_config


def test_with_feature(tmpdir, env_feature_config):
    conf = env_feature_config(["test"])
    test = tmpdir.join("test.py")
    test.write(
        """
import pytest

@pytest.mark.lg_feature("test")
def test(env):
    assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 passed")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


def test_skip_feature(tmpdir, env_feature_config):
    conf = env_feature_config(["test"])
    test = tmpdir.join("test2.py")
    test.write(
        """
import pytest

@pytest.mark.lg_feature("test2")
def test(env):
    assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 skipped")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


def test_skip_feature_list(tmpdir, env_feature_config):
    conf = env_feature_config(["test"])
    test = tmpdir.join("test2.py")
    test.write(
        """
import pytest

@pytest.mark.lg_feature(["test2", "test3"])
def test(env):
    assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 skipped")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


def test_match_feature_list(tmpdir, env_feature_config):
    conf = env_feature_config(["test1", "test2"])
    test = tmpdir.join("test2.py")
    test.write(
        """
import pytest

@pytest.mark.lg_feature(["test1", "test2"])
def test(env):
    assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 passed")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


def test_match_multi_feature_source(tmpdir, env_feature_config):
    conf = env_feature_config(["test1", "test2", "test3"])
    test = tmpdir.join("test.py")
    test.write(
        """
import pytest

pytestmark = pytest.mark.lg_feature("test1")

@pytest.mark.lg_feature("test2")
class TestMulti:
    @pytest.mark.lg_feature("test3")
    def test(self, env):
        assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 passed")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


def test_skip_multi_feature_source(tmpdir, env_feature_config):
    conf = env_feature_config(["test1", "test3"])
    test = tmpdir.join("test.py")
    test.write(
        """
import pytest

pytestmark = pytest.mark.lg_feature("test1")

@pytest.mark.lg_feature("test2")
class TestMulti:
    @pytest.mark.lg_feature("test3")
    def test(self, env):
        assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 skipped")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0


@pytest.mark.parametrize(
    "marker_args_str,error",
    [
        ("", "Unexpected number of args/kwargs"),
        ("'too', 'many'", "Unexpected number of args/kwargs"),
        ("{'foo': 'bar'}", "Unsupported 'features' argument type"),
    ],
    ids=["no args", "too many args", "unsupported arg type"],
)
def test_lg_feature_unexpected_args(tmpdir, env_feature_config, marker_args_str, error):
    # features do not matter here, simply generate a valid env config
    conf = env_feature_config([])
    test = tmpdir.join("test.py")
    test.write(
        f"""
import pytest

@pytest.mark.lg_feature({marker_args_str})
def test(self, env):
    assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect(error)
        spawn.expect(pexpect.EOF)
        spawn.close()
        # pytest command line usage error leads to exit code 4
        assert spawn.exitstatus == 4


def test_xfail_feature(tmpdir, env_feature_config):
    conf = env_feature_config(["test"])
    test = tmpdir.join("test.py")
    test.write(
        """
import pytest

@pytest.mark.lg_xfail_feature("test")
def test(env):
    assert False
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 xfailed")
        spawn.expect(pexpect.EOF)
    assert spawn.exitstatus == 0


def test_no_xfail_feature(tmpdir, env_feature_config):
    conf = env_feature_config([])
    test = tmpdir.join("test.py")
    test.write(
        """
import pytest

@pytest.mark.lg_xfail_feature("test")
def test(env):
    assert False
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect("1 failed")
        spawn.expect(pexpect.EOF)
    assert spawn.exitstatus == 1


@pytest.mark.parametrize(
    "marker_args_str,error",
    [
        ("", "Unexpected number of arguments"),
        ("'too', 'many'", "Unexpected number of arguments"),
        ("{'foo': 'bar'}", "Unsupported 'feature' argument type"),
        ("'feature', condition='mycondition'", "Unsupported 'condition' argument"),
    ],
    ids=["no args", "too many args", "unsupported arg type", "unsupported condition"],
)
def test_lg_xfail_feature_unexpected_args(tmpdir, env_feature_config, marker_args_str, error):
    # features do not matter here, simply generate a valid env config
    conf = env_feature_config([])
    test = tmpdir.join("test.py")
    test.write(
        f"""
import pytest

@pytest.mark.lg_xfail_feature({marker_args_str})
def test(self, env):
    assert True
"""
    )

    with pexpect.spawn(f"pytest --lg-env {conf} {test}") as spawn:
        spawn.expect(error)
        spawn.expect(pexpect.EOF)
        spawn.close()
        # pytest command line usage error leads to exit code 4
        assert spawn.exitstatus == 4



from .target import Target
from .environment import Environment
from .exceptions import NoConfigFoundError

from .factory import target_factory
from .step import step, steps
from .stepreporter import StepReporter
from .consoleloggingreporter import ConsoleLoggingReporter

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"



import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..step import step
from .common import Driver
from . import SerialDriver


@target_factory.reg_driver
@attr.s(eq=False)
class SerialPortDigitalOutputDriver(Driver, DigitalOutputProtocol):
    """
    Controls the state of a GPIO using the control lines of a serial port.

    This driver uses the flow-control pins of a serial port (for example
    an USB-UART-dongle) to control some external power switch. You may connect
    some kind of relay board to the flow control pins.

    The serial port should NOT be used for serial communication at the same
    time. This will probably reset the flow-control signals.

    Usable signals are DTR and RTS.
    """

    bindings = {"serial": SerialDriver}
    signal = attr.ib(validator=attr.validators.instance_of(str))
    invert = attr.ib(validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        # basic input format checking
        self.signal = self.signal.lower()
        if self.signal not in ["dtr", "rts"]:
            raise Exception("Unknown index for serial-power-driver")

        self._p = self.serial.serial

    @Driver.check_active
    @step()
    def get(self):
        if self.signal == "dtr":
            val = self._p.dtr
        elif self.signal == "rts":
            val = self._p.rts
        else:
            raise ValueError("Expected signal to be dtr or rts")

        if self.invert:
            val = not val

        return val

    @Driver.check_active
    @step()
    def set(self, value):
        if self.invert:
            value = not value
        if self.signal == "dtr":
            self._p.dtr = value
        elif self.signal == "rts":
            self._p.rts = value



import abc


class LinuxBootProtocol(abc.ABC):
    @abc.abstractmethod
    def boot(self, name: str):
        raise NotImplementedError

    @abc.abstractmethod
    def await_boot(self):
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self):
        raise NotImplementedError



import attr

from ..factory import target_factory
from .common import Resource
from .base import SerialPort


@target_factory.reg_resource
@attr.s(eq=False)
class ModbusRTU(SerialPort, Resource):
    """This resource describes Modbus RTU instrument.

    Args:
        port (str): tty the instrument is connected to, e.g. '/dev/ttyUSB0'
        speed (int): optional, default is 115200
        address (int): slave address on the modbus, e.g. 16
        timeout (float): optional, timeout in seconds. Default is 0.25 s
    """

    address = attr.ib(default=None, validator=attr.validators.instance_of(int))
    timeout = attr.ib(default=0.25,
                      validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.port is None:
            raise ValueError("ModbusRTU must be configured with a port")
        if self.address is None:
            raise ValueError("ModbusRTU must be configured with an slave address")



from collections import OrderedDict

import pytest

from labgrid.config import Config
from labgrid.exceptions import InvalidConfigError


def test_get_target_option(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        targets:
          main:
            options:
              str: test
              list: [1, 2, 3]
              dict:
                a: 1
                b: 2
              bool: False
              int: 0x20
              float: 3.14
              none: null
        """
    )
    c = Config(str(p))
    assert c.get_target_option("main", "str") == "test"
    assert c.get_target_option("main", "list") == [1, 2, 3]
    assert c.get_target_option("main", "dict") == OrderedDict([("a", 1), ("b", 2)])
    assert c.get_target_option("main", "bool") is False
    assert c.get_target_option("main", "int") == 0x20
    assert c.get_target_option("main", "float") == 3.14
    assert c.get_target_option("main", "none") is None

    with pytest.raises(KeyError) as err:
        c.get_target_option("main", "blah")
    assert "No option" in str(err)

    with pytest.raises(KeyError) as err:
        c.get_target_option("nonexist", "str")
    assert "No target" in str(err)


def test_set_target_option(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        targets:
          main:
        """
    )
    c = Config(str(p))

    with pytest.raises(KeyError) as err:
        c.get_target_option("main", "spam")
    assert "No option" in str(err)

    c.set_target_option("main", "spam", "eggs")
    assert c.get_target_option("main", "spam") == "eggs"

    obj = object()
    c.set_target_option("main", "obj", obj)
    assert c.get_target_option("main", "obj") is obj


def test_template(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        dict:
          list:
          - a
          - b
          - !template $BASE
          string: !template ${BASE}/suffix
        """
    )
    c = Config(str(p))
    assert "a" in c.data["dict"]["list"]
    assert c.data["dict"]["list"][2] == str(tmpdir)
    assert c.data["dict"]["string"] == str(tmpdir) + "/suffix"


def test_template_bad_placeholder(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        string: !template $
        """
    )
    with pytest.raises(InvalidConfigError) as excinfo:
        Config(str(p))
    assert "is invalid" in excinfo.value.msg
    assert "template string" in excinfo.value.msg


def test_template_bad_key(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
        """
        string: !template ${INVALID}
        """
    )
    with pytest.raises(InvalidConfigError) as excinfo:
        Config(str(p))
    assert "unknown variable" in excinfo.value.msg


def test_tool(tmpdir):
    t = tmpdir.join("testtool")
    t.write("content")
    p = tmpdir.join("config.yaml")
    p.write(
        """
        tools:
          testtool: {}
        """.format(t)
    )
    c = Config(str(p))

    assert c.get_tool("testtool") == t


def test_tool_no_explicit_tool(tmpdir):
    t = tmpdir.join("testtool")
    t.write("content")
    p = tmpdir.join("config.yaml")
    p.write(
        """
        dict: {}
        """
    )
    c = Config(str(p))

    assert c.get_tool("testtool") == "testtool"



"""tested with NETIO 4C, should be compatible with all NETIO 4-models"""

import re

import pexpect

PORT = 1234


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 4
    value = "1" if value else "0"

    with pexpect.spawn(f"telnet {host} {port}", timeout=1) as tn:
        tn.expect(b"100 HELLO .*\r\n")
        tn.send(b"login admin admin\r\n")

        tn.expect(b"250 OK\r\n")
        tn.send(f"port {index} {value}\r\n".encode())

        tn.expect(b"250 OK\r\n")
        tn.send(b"quit\r\n")
        tn.expect(pexpect.EOF)


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 4

    with pexpect.spawn(f"telnet {host} {port}", timeout=1) as tn:
        tn.expect(b"100 HELLO .*\r\n")
        tn.send(b"login admin admin\r\n")

        tn.expect(b"250 OK\r\n")
        tn.send(f"port {index}\r\n".encode())

        tn.expect(rb"250 .*\r\n")
        m = re.match(r".*250 (\d).*", tn.after.decode())
        if m is None:
            raise Exception("NetIO: could not match response")
        value = m.group(1)

        tn.send(b"quit\r\n")
        tn.expect(pexpect.EOF)

    return value == "1"



import abc


class DigitalOutputProtocol(abc.ABC):
    """Abstract class providing the DigitalOutputProtocol interface"""

    @abc.abstractmethod
    def get(self):
        """Implementations should return the status of the digital output."""
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, status):
        """Implementations should set the status of the digital output"""
        raise NotImplementedError



import pytest


def test_state(command):
    stdout, stderr, returncode = command.run("state")
    if returncode == 127:
        pytest.skip("state command not available")
    assert returncode == 0
    assert not stderr
    assert stdout[0] == "registered state instances:"
    assert len(stdout) > 1



from collections import OrderedDict, UserString

import pytest
import yaml

from labgrid.util.yaml import *


def test_default_loader():
    """Importing labgrid should not modify the default loaders"""
    doc = """
    foo: |
      multi
      line
    bar: False
    """

    data = yaml.load(doc, Loader=yaml.SafeLoader)
    assert type(data) == dict
    assert type(data["foo"]) == str

    data = yaml.load(doc, Loader=yaml.FullLoader)
    assert type(data) == dict
    assert type(data["foo"]) == str

    data = yaml.load(doc, Loader=yaml.Loader)
    assert type(data) == dict
    assert type(data["foo"]) == str


def test_labgrid_loader():
    doc = """
    foo: |
      multi
      line
    bar: False
    """
    data = load(doc)
    assert type(data) == OrderedDict
    assert type(data["foo"]) == UserString
    assert data["foo"].start_mark.line == 1
    assert data["foo"].end_mark.line == 4


def test_default_dumper():
    """Importing labgrid should not modify the default dumpers"""
    data = OrderedDict([])
    assert "!!python/object/apply:collections.OrderedDict\n- []\n" == yaml.dump(data, Dumper=yaml.Dumper)
    with pytest.raises(yaml.representer.RepresenterError):
        yaml.dump(data, Dumper=yaml.SafeDumper)


def test_labgrid_dumper():
    data = OrderedDict([])
    doc = dump(data)
    assert "{}\n" == doc



import pytest

from labgrid import Target
from labgrid.driver import UBootDriver
from labgrid.driver.fake import FakeConsoleDriver
from labgrid.protocol import CommandProtocol, LinuxBootProtocol
from labgrid.driver import ExecutionError


class TestUBootDriver:
    def test_create(self):
        t = Target('dummy')
        cp = FakeConsoleDriver(t, "console")
        d = UBootDriver(t, "uboot")
        assert (isinstance(d, UBootDriver))
        assert (isinstance(d, CommandProtocol))
        assert (isinstance(d, LinuxBootProtocol))

    def test_uboot_run(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = UBootDriver(t, "uboot")
        d.console.rxq.append(b"U-Boot 2019\n")
        d = t.get_driver(UBootDriver)
        d._run = mocker.MagicMock(return_value=(['success'], [], 0))
        res = d.run_check("test")
        assert res == ['success']
        d._run.assert_called_once_with("test", timeout=30, codec='utf-8', decodeerrors='strict')
        d._run.reset_mock()
        res = d.run("test")
        assert res == (['success'], [], 0)
        d._run.assert_called_once_with("test", timeout=30)

    def test_uboot_run_error(self, target_with_fakeconsole, mocker):
        t = target_with_fakeconsole
        d = UBootDriver(t, "uboot")
        d.console.rxq.append(b"U-Boot 2019\n")
        d = t.get_driver(UBootDriver)
        d._run = mocker.MagicMock(return_value=(['error'], [], 1))
        with pytest.raises(ExecutionError):
            res = d.run_check("test")
        d._run.assert_called_once_with("test", timeout=30, codec='utf-8', decodeerrors='strict')
        d._run.reset_mock()
        res = d.run("test")
        assert res == (['error'], [], 1)
        d._run.assert_called_once_with("test", timeout=30)

    def test_uboot_boot(self, target_with_fakeconsole):
        t = target_with_fakeconsole
        d = UBootDriver(t, "uboot", boot_command='run bootcmd', boot_commands = {"foo": "bar"})
        d = t.get_driver(UBootDriver)
        d.boot()
        assert d.console.txq.pop() == b"run bootcmd\n"
        d.boot('foo')
        assert d.console.txq.pop() == b"bar\n"
        with pytest.raises(Exception):
            d.boot('nop')



import os.path
import attr

from ..factory import target_factory
from ..step import step
from .common import Driver
from ..util.managedfile import ManagedFile


@attr.s(eq=False)
class BaseProviderDriver(Driver):
    @Driver.check_bound
    def get_export_vars(self):
        return {
            "host": self.provider.host,
            "internal": self.provider.internal,
            "external": self.provider.external,
        }

    @Driver.check_active
    @step(args=["filename"], result=True)
    def stage(self, filename):
        symlink = os.path.join(self.provider.internal, os.path.basename(filename))
        assert symlink.startswith(self.provider.internal)

        mf = ManagedFile(filename, self.provider)
        mf.sync_to_resource(symlink=symlink)

        return self.provider.external + symlink[len(self.provider.internal) :]


@target_factory.reg_driver
@attr.s(eq=False)
class TFTPProviderDriver(BaseProviderDriver):
    bindings = {
        "provider": {"TFTPProvider", "RemoteTFTPProvider"},
    }


@attr.s
class NFSFile:
    host = attr.ib(validator=attr.validators.instance_of(str))
    export = attr.ib(validator=attr.validators.instance_of(str))
    relative_file_path = attr.ib(validator=attr.validators.instance_of(str))


@target_factory.reg_driver
@attr.s(eq=False)
class NFSProviderDriver(Driver):
    bindings = {
        "provider": {"NFSProvider", "RemoteNFSProvider"},
    }

    @Driver.check_bound
    def get_export_vars(self):
        return {
            "host": self.provider.host,
        }

    @Driver.check_active
    @step(args=["filename"], result=True)
    def stage(self, filename):
        # always copy the file to he user cache path:
        # locally available files might not be NFS-exported
        mf = ManagedFile(filename, self.provider, detect_nfs=False)
        mf.sync_to_resource()
        mf.get_remote_path()

        # assuming /var/cache/labgrid is NFS-exported, return required information for mounting and
        # file access
        relate_file_path = os.path.join(mf.get_hash(), os.path.basename(mf.local_path))
        return NFSFile(self.provider.host, mf.get_user_cache_path(), relate_file_path)


@target_factory.reg_driver
@attr.s(eq=False)
class HTTPProviderDriver(BaseProviderDriver):
    bindings = {
        "provider": {"HTTPProvider", "RemoteHTTPProvider"},
    }



#!/usr/bin/env python3
import argparse
import logging
import asyncio
import traceback
from enum import Enum
from functools import wraps
import time
from contextlib import contextmanager
import copy
import random
import signal

import attr
import grpc
from grpc_reflection.v1alpha import reflection

from .common import (
    ResourceEntry,
    ResourceMatch,
    Place,
    Reservation,
    ReservationState,
    queue_as_aiter,
    TAG_KEY,
    TAG_VAL,
)
from .scheduler import TagSet, schedule
from .generated import labgrid_coordinator_pb2
from .generated import labgrid_coordinator_pb2_grpc
from ..util import atomic_replace, labgrid_version, yaml, Timeout


@contextmanager
def warn_if_slow(prefix, *, level=logging.WARNING, limit=0.1):
    monotonic = time.monotonic()
    process = time.process_time()
    thread = time.thread_time()
    yield
    monotonic = time.monotonic() - monotonic
    process = time.process_time() - process
    thread = time.thread_time() - thread
    if monotonic > limit:
        logging.log(level, "%s: real %.3f>%.3f, process %.3f, thread %.3f", prefix, monotonic, limit, process, thread)


class Action(Enum):
    ADD = 0
    DEL = 1
    UPD = 2


@attr.s(init=False, eq=False)
class RemoteSession:
    """class encapsulating a session, used by ExporterSession and ClientSession"""

    coordinator = attr.ib()
    peer = attr.ib()
    name = attr.ib()
    queue = attr.ib()
    version = attr.ib()


@attr.s(eq=False)
class ExporterSession(RemoteSession):
    """An ExporterSession is opened for each Exporter connecting to the
    coordinator, allowing the Exporter to get and set resources"""

    groups = attr.ib(default=attr.Factory(dict), init=False)

    def set_resource(self, groupname, resourcename, resource):
        """This is called when Exporters update resources or when they disconnect."""
        logging.info("set_resource %s %s %s", groupname, resourcename, resource)
        group = self.groups.setdefault(groupname, {})
        old: ResourceImport = group.get(resourcename)
        if resource is not None:
            new = ResourceImport(
                data=ResourceImport.data_from_pb2(resource), path=(self.name, groupname, resource.cls, resourcename)
            )
            if old:
                old.data.update(new.data)
                new = old
            else:
                group[resourcename] = new
        else:
            new = None
            if old.acquired:
                old.orphaned = True
            try:
                del group[resourcename]
            except KeyError:
                pass

        msg = labgrid_coordinator_pb2.ClientOutMessage()
        update = msg.updates.add()
        if new:
            update.resource.CopyFrom(new.as_pb2())
            update.resource.path.exporter_name = self.name
            update.resource.path.group_name = groupname
            update.resource.path.resource_name = resourcename
        else:
            update.del_resource.exporter_name = self.name
            update.del_resource.group_name = groupname
            update.del_resource.resource_name = resourcename

        for client in self.coordinator.clients.values():
            client.queue.put_nowait(msg)

        if old and new:
            assert old is new
            return Action.UPD, new
        elif old and not new:
            return Action.DEL, old
        elif not old and new:
            return Action.ADD, new

        assert not old and not new

    def get_resources(self):
        """Method invoked by the client, get the resources from the coordinator"""
        result = {}
        for groupname, group in self.groups.items():
            result_group = result[groupname] = {}
            for resourcename, resource in group.items():
                result_group[resourcename] = resource.asdict()
        return result


@attr.s(eq=False)
class ClientSession(RemoteSession):
    def subscribe_places(self):
        # send initial places
        out_msg = labgrid_coordinator_pb2.ClientOutMessage()
        for place in self.coordinator.places.values():
            place: Place
            out_msg.updates.add().place.CopyFrom(place.as_pb2())
        self.queue.put_nowait(out_msg)

    def subscribe_resources(self):
        # collect initial resources
        collected = []
        logging.debug("sending resources to %s", self)
        for exporter in self.coordinator.exporters.values():
            logging.debug("sending resources %s", exporter)
            exporter: ExporterSession
            for groupname, group in exporter.groups.items():
                logging.debug("sending resources %s", groupname)
                for resourcename, resource in group.items():
                    logging.debug("sending resources %s", resourcename)
                    resource: ResourceImport
                    update = labgrid_coordinator_pb2.UpdateResponse()
                    update.resource.CopyFrom(resource.as_pb2())
                    update.resource.path.exporter_name = exporter.name
                    update.resource.path.group_name = groupname
                    update.resource.path.resource_name = resourcename
                    collected.append(update)
        # send batches
        while collected:
            batch, collected = collected[:100], collected[100:]
            out_msg = labgrid_coordinator_pb2.ClientOutMessage()
            out_msg.updates.extend(batch)
            self.queue.put_nowait(out_msg)


@attr.s(eq=False)
class ResourceImport(ResourceEntry):
    """Represents a local resource exported from an exporter.

    The ResourceEntry attributes contain the information for the client.
    """

    path = attr.ib(kw_only=True, validator=attr.validators.instance_of(tuple))
    orphaned = attr.ib(init=False, default=False, validator=attr.validators.instance_of(bool))


def locked(func):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        async with self.lock:
            return await func(self, *args, **kwargs)

    return wrapper


class ExporterCommand:
    def __init__(self, request) -> None:
        self.request = request
        self.response = None
        self.completed = asyncio.Event()
        self.expired = False

    def complete(self, response) -> None:
        self.response = response
        self.completed.set()
        if self.expired:
            logging.warning(
                "exporter command already expired for request %s -> response %s", self.request, self.response
            )

    async def wait(self):
        try:
            await asyncio.wait_for(self.completed.wait(), 10)
        finally:
            if self.response is None:
                self.expired = True


class ExporterError(Exception):
    pass


class Coordinator(labgrid_coordinator_pb2_grpc.CoordinatorServicer):
    def __init__(self) -> None:
        self.places: dict[str, Place] = {}
        self.reservations = {}
        self.poll_tasks = []
        self.save_scheduled = False

        self.lock = asyncio.Lock()
        self.exporters: dict[str, ExporterSession] = {}
        self.clients: dict[str, ClientSession] = {}
        self.load()

        self.loop = asyncio.get_running_loop()
        for name in ["save", "sync_resources", "schedule"]:
            step_func = getattr(self, f"_poll_step_{name}")
            task = self.loop.create_task(self.poll(step_func), name=f"coordinator-poll-{name}")
            self.poll_tasks.append(task)

    async def _poll_step_save(self):
        # save changes
        if self.save_scheduled:
            with warn_if_slow("save changes", level=logging.DEBUG):
                await self.save()

    async def _poll_step_sync_resources(self):
        # try to synchronize resources
        async with self.lock:
            with warn_if_slow("synchronize resources", limit=3.0):
                await self._synchronize_resources()

    async def _poll_step_schedule(self):
        # update reservations
        async with self.lock:
            with warn_if_slow("schedule reservations"):
                self.schedule_reservations()

    async def poll(self, step_func):
        while not self.loop.is_closed():
            try:
                await asyncio.sleep(15.0)
                await step_func()
            except asyncio.CancelledError:
                break
            except Exception:  # pylint: disable=broad-except
                traceback.print_exc()

    def save_later(self):
        logging.debug("Setting Save-later")
        self.save_scheduled = True

    def _get_resources(self):
        result = {}
        for session in self.exporters.values():
            result[session.name] = session.get_resources()
        return result

    async def save(self):
        logging.debug("Running Save")
        self.save_scheduled = False

        with warn_if_slow("create resources snapshot", level=logging.DEBUG):
            resources = copy.deepcopy(self._get_resources())
        with warn_if_slow("create places snapshot", level=logging.DEBUG):
            places = copy.deepcopy(self._get_places())

        def save_sync(resources, places):
            with warn_if_slow("dump resources", level=logging.DEBUG):
                resources = yaml.dump(resources)
                resources = resources.encode()
            with warn_if_slow("dump places", level=logging.DEBUG):
                places = yaml.dump(places)
                places = places.encode()
            with warn_if_slow("write resources", level=logging.DEBUG):
                atomic_replace("resources.yaml", resources)
            with warn_if_slow("write places", level=logging.DEBUG):
                atomic_replace("places.yaml", places)

        await self.loop.run_in_executor(None, save_sync, resources, places)

    def load(self):
        try:
            self.places = {}
            with open("places.yaml", "r") as f:
                self.places = yaml.load(f.read())
            for placename, config in self.places.items():
                config["name"] = placename
                # FIXME maybe recover previously acquired places here?
                if "acquired" in config:
                    del config["acquired"]
                if "acquired_resources" in config:
                    del config["acquired_resources"]
                if "allowed" in config:
                    del config["allowed"]
                if "reservation" in config:
                    del config["reservation"]
                config["matches"] = [ResourceMatch(**match) for match in config["matches"]]
                place = Place(**config)
                self.places[placename] = place
        except FileNotFoundError:
            pass
        logging.info("loaded %s place(s)", len(self.places))

    async def ClientStream(self, request_iterator, context):
        peer = context.peer()
        logging.info("client connected: %s", peer)
        assert peer not in self.clients
        out_msg_queue = asyncio.Queue()

        async def request_task():
            name = None
            version = None
            try:
                async for in_msg in request_iterator:
                    in_msg: labgrid_coordinator_pb2.ClientInMessage
                    logging.debug("client in_msg %s", in_msg)
                    kind = in_msg.WhichOneof("kind")
                    if kind == "sync":
                        out_msg = labgrid_coordinator_pb2.ClientOutMessage()
                        out_msg.sync.id = in_msg.sync.id
                        out_msg_queue.put_nowait(out_msg)
                    elif kind == "startup":
                        version = in_msg.startup.version
                        name = in_msg.startup.name
                        session = self.clients[peer] = ClientSession(self, peer, name, out_msg_queue, version)
                        logging.debug("Received startup from %s with %s", name, version)
                        asyncio.current_task().set_name(f"client-{peer}-rx/started-{name}")
                    elif kind == "subscribe":
                        if in_msg.subscribe.all_places:
                            session.subscribe_places()
                        if in_msg.subscribe.all_resources:
                            session.subscribe_resources()
                    else:
                        logging.warning("received unknown kind %s from client %s (version %s)", kind, name, version)
                logging.debug("client request_task done: %s", context.done())
            except Exception:
                logging.exception("error in client message handler")

        asyncio.current_task().set_name(f"client-{peer}-tx")
        running_request_task = self.loop.create_task(request_task(), name=f"client-{peer}-rx/init")

        try:
            async for out_msg in queue_as_aiter(out_msg_queue):
                out_msg: labgrid_coordinator_pb2.ClientOutMessage
                logging.debug("client output %s", out_msg)
                yield out_msg
        finally:
            try:
                session = self.clients.pop(peer)
                running_request_task.cancel()
                await running_request_task
                logging.debug("client aborted %s, cancelled: %s", session, context.cancelled())
            except KeyError:
                logging.info("Never received startup from peer %s that disconnected", peer)

    def _add_default_place(self, name):
        if name in self.places:
            return
        if not name.isdigit():
            return
        place = Place(name)
        print(place)
        place.matches.append(ResourceMatch(exporter="*", group=name, cls="*"))
        self.places[name] = place

    def get_exporter_by_name(self, name):
        for exporter in self.exporters.values():
            if exporter.name == name:
                return exporter

    def _publish_place(self, place):
        msg = labgrid_coordinator_pb2.ClientOutMessage()
        msg.updates.add().place.CopyFrom(place.as_pb2())

        for client in self.clients.values():
            client.queue.put_nowait(msg)

    def _publish_resource(self, resource: ResourceImport):
        msg = labgrid_coordinator_pb2.ClientOutMessage()
        update = msg.updates.add()
        update.resource.CopyFrom(resource.as_pb2())
        update.resource.path.exporter_name = resource.path[0]
        update.resource.path.group_name = resource.path[1]
        update.resource.path.resource_name = resource.path[3]

        for client in self.clients.values():
            client.queue.put_nowait(msg)

    async def ExporterStream(self, request_iterator, context):
        peer = context.peer()
        logging.info("exporter connected: %s", peer)
        assert peer not in self.exporters
        command_queue = asyncio.Queue()
        pending_commands = []

        startup_done = asyncio.Event()

        out_msg = labgrid_coordinator_pb2.ExporterOutMessage()
        out_msg.hello.version = labgrid_version()
        yield out_msg

        async def request_task():
            name = None
            version = None
            try:
                async for in_msg in request_iterator:
                    in_msg: labgrid_coordinator_pb2.ExporterInMessage
                    logging.debug("exporter in_msg %s", in_msg)
                    kind = in_msg.WhichOneof("kind")
                    if kind in "response":
                        cmd = pending_commands.pop(0)
                        cmd.complete(in_msg.response)
                        logging.debug("Command %s is done", cmd)
                    elif kind == "startup":
                        version = in_msg.startup.version
                        name = in_msg.startup.name
                        if existing := self.get_exporter_by_name(name):
                            raise ExporterError(
                                f"exporter with name '{name}' is already connected from {existing.peer}"
                            )
                        session = self.exporters[peer] = ExporterSession(self, peer, name, command_queue, version)
                        logging.debug("Exporters: %s", self.exporters)
                        logging.debug("Received startup from %s with %s", name, version)
                        asyncio.current_task().set_name(f"exporter-{peer}-rx/started-{name}")
                        startup_done.set()
                    elif kind == "resource":
                        logging.debug("Received resource from %s with %s", name, in_msg.resource)
                        action, _ = session.set_resource(
                            in_msg.resource.path.group_name, in_msg.resource.path.resource_name, in_msg.resource
                        )
                        if action is Action.ADD:
                            async with self.lock:
                                self._add_default_place(in_msg.resource.path.group_name)
                        self.save_later()
                    else:
                        logging.warning("received unknown kind %s from exporter %s (version %s)", kind, name, version)

                logging.debug("exporter request_task done: %s", context.done())
            except Exception:
                logging.exception("error in exporter message handler")
                raise

        asyncio.current_task().set_name(f"exporter-{peer}-tx")
        running_request_task = self.loop.create_task(request_task(), name=f"exporter-{peer}-rx/init")

        startup_done_task = self.loop.create_task(startup_done.wait())
        done, _ = await asyncio.wait(
            {startup_done_task, running_request_task},
            timeout=3,
            return_when=asyncio.FIRST_COMPLETED,
        )
        # clean up event task
        startup_done.set()
        await startup_done_task
        if running_request_task in done:
            # we probably had an exception during startup
            try:
                await running_request_task
            except ExporterError as e:
                await context.abort(grpc.StatusCode.ALREADY_EXISTS, f"startup failed: {e}")
                raise
        elif startup_done_task in done:
            await startup_done_task
        else:
            raise ExporterError(f"exporter connection from {peer} timed out during startup")

        try:
            async for cmd in queue_as_aiter(command_queue):
                logging.debug("exporter cmd %s", cmd)
                out_msg = labgrid_coordinator_pb2.ExporterOutMessage()
                out_msg.set_acquired_request.CopyFrom(cmd.request)
                pending_commands.append(cmd)
                yield out_msg
        except asyncio.exceptions.CancelledError:
            logging.info("exporter disconnected %s", context.peer())
        except Exception:
            logging.exception("error in exporter command handler")
        finally:
            running_request_task.cancel()
            await running_request_task

            try:
                session = self.exporters.pop(peer)
                for groupname, group in session.groups.items():
                    for resourcename in group.copy():
                        session.set_resource(groupname, resourcename, None)

                logging.debug("exporter aborted %s, cancelled: %s", context.peer(), context.cancelled())
            except KeyError:
                logging.info("Never received startup from peer %s that disconnected", peer)

    @locked
    async def AddPlace(self, request, context):
        name = request.name
        if not name or not isinstance(name, str):
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, "name was not a string")
        if name in self.places:
            await context.abort(grpc.StatusCode.ALREADY_EXISTS, f"Place {name} already exists")
        logging.debug("Adding %s", name)
        place = Place(name)
        self.places[name] = place
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.AddPlaceResponse()

    @locked
    async def DeletePlace(self, request, context):
        name = request.name
        if not name or not isinstance(name, str):
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, "name was not a string")
        if name not in self.places:
            await context.abort(grpc.StatusCode.ALREADY_EXISTS, f"Place {name} does not exist")
        logging.debug("Deleting %s", name)
        del self.places[name]
        msg = labgrid_coordinator_pb2.ClientOutMessage()
        msg.updates.add().del_place = name
        for client in self.clients.values():
            client.queue.put_nowait(msg)
        self.save_later()
        return labgrid_coordinator_pb2.DeletePlaceResponse()

    @locked
    async def AddPlaceAlias(self, request, context):
        placename = request.placename
        alias = request.alias
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        place.aliases.add(alias)
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.AddPlaceAliasResponse()

    @locked
    async def DeletePlaceAlias(self, request, context):
        placename = request.placename
        alias = request.alias
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        try:
            place.aliases.remove(alias)
        except ValueError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Failed to remove {alias} from {placename}")
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.DeletePlaceAliasResponse()

    @locked
    async def SetPlaceTags(self, request, context):
        placename = request.placename
        tags = dict(request.tags)
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        assert isinstance(tags, dict)
        for k, v in tags.items():
            assert isinstance(k, str)
            assert isinstance(v, str)
            if not TAG_KEY.match(k):
                await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Key {k} in {tags} is invalid")
            if not TAG_VAL.match(v):
                await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Value {v} in {tags} is invalid")
        for k, v in tags.items():
            if not v:
                try:
                    del place.tags[k]
                except KeyError:
                    pass
            else:
                place.tags[k] = v
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.SetPlaceTagsResponse()

    @locked
    async def SetPlaceComment(self, request, context):
        placename = request.placename
        comment = request.comment
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        place.comment = comment
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.SetPlaceCommentResponse()

    @locked
    async def AddPlaceMatch(self, request, context):
        placename = request.placename
        pattern = request.pattern
        rename = request.rename if request.HasField("rename") else None
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        rm = ResourceMatch(*pattern.split("/"), rename=rename)
        if rm in place.matches:
            await context.abort(grpc.StatusCode.ALREADY_EXISTS, f"Match {rm} already exists")
        place.matches.append(rm)
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.AddPlaceMatchResponse()

    @locked
    async def DeletePlaceMatch(self, request, context):
        placename = request.placename
        pattern = request.pattern
        rename = request.rename if request.HasField("rename") else None
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        rm = ResourceMatch(*pattern.split("/"), rename=rename)
        try:
            place.matches.remove(rm)
        except ValueError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Match {rm} does not exist in {placename}")
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.DeletePlaceMatchResponse()

    async def _acquire_resource(self, place, resource):
        assert self.lock.locked()

        # this triggers an update from the exporter which is published
        # to the clients
        request = labgrid_coordinator_pb2.ExporterSetAcquiredRequest()
        request.group_name = resource.path[1]
        request.resource_name = resource.path[3]
        request.place_name = place.name
        cmd = ExporterCommand(request)
        self.get_exporter_by_name(resource.path[0]).queue.put_nowait(cmd)
        await cmd.wait()
        if not cmd.response.success:
            raise ExporterError(f"failed to acquire {resource} ({cmd.response.reason})")
        if resource.acquired != place.name:
            logging.warning("resource %s not acquired by this place after acquire request", resource)

    async def _acquire_resources(self, place, resources):
        assert self.lock.locked()

        resources = resources.copy()  # we may modify the list
        # all resources need to be free
        for resource in resources:
            if resource.acquired:
                return False

            for otherplace in self.places.values():
                for oldres in otherplace.acquired_resources:
                    if resource.path == oldres.path:
                        logging.info(
                            "Conflicting orphaned resource %s for acquire request for place %s", oldres, place.name
                        )
                        return False

        # acquire resources
        acquired = []
        try:
            for resource in resources:
                await self._acquire_resource(place, resource)
                acquired.append(resource)
        except Exception:
            logging.exception("failed to acquire %s", resource)
            # cleanup
            await self._release_resources(place, acquired)
            return False

        for resource in resources:
            place.acquired_resources.append(resource)

        return True

    async def _release_resources(self, place, resources, callback=True):
        assert self.lock.locked()

        resources = resources.copy()  # we may modify the list

        for resource in resources:
            try:
                place.acquired_resources.remove(resource)
            except ValueError:
                pass

        for resource in resources:
            if resource.orphaned:
                continue
            try:
                # this triggers an update from the exporter which is published
                # to the clients
                if callback:
                    request = labgrid_coordinator_pb2.ExporterSetAcquiredRequest()
                    request.group_name = resource.path[1]
                    request.resource_name = resource.path[3]
                    # request.place_name is left unset to indicate release
                    cmd = ExporterCommand(request)
                    self.get_exporter_by_name(resource.path[0]).queue.put_nowait(cmd)
                    await cmd.wait()
                    if not cmd.response.success:
                        raise ExporterError(f"failed to release {resource} ({cmd.response.reason})")
                    if resource.acquired:
                        logging.warning("resource %s still acquired after release request", resource)
            except (ExporterError, TimeoutError):
                logging.exception("failed to release %s", resource)
                # at leaset try to notify the clients
                try:
                    self._publish_resource(resource)
                except:
                    logging.exception("failed to publish released resource %s", resource)

    async def _synchronize_resources(self):
        assert self.lock.locked()

        # fix:
        # - a resource is acquired for a place that is not acquired
        #   * perhaps caused by a resource acquire timeout (during normal lock)
        #   -> release()
        # - a resource is acquired for a place that still has it as orphaned
        #   * perhaps caused by a resource acquire timeout (during reacquire)
        #   -> replace orphaned resource
        # - a resource is released, but a place still has it as orphaned
        #   * perhaps caused by a exporter restart
        #   -> acquire() and replace orphaned resource

        acquired_resources = {}
        used_resources = {}
        orphaned_resources = {}

        # find acquired resources
        for exporter in self.exporters.values():
            for group in exporter.groups.values():
                for resource in group.values():
                    if resource.acquired:
                        acquired_resources[resource.path] = resource

        # find resources used by places
        for place in self.places.values():
            for resource in place.acquired_resources:
                if not resource.orphaned:
                    used_resources[resource.path] = resource
                else:
                    orphaned_resources[resource.path] = resource

        timeout = Timeout(5.0)

        # find resources to be released
        to_release = list(acquired_resources.keys() - used_resources.keys() - orphaned_resources.keys())
        if to_release:
            logging.info("synchronize resources: %s acquired resource(s) should be released", len(to_release))
            random.shuffle(to_release)  # don't get stuck on a problematic resource
        for resource_path in to_release:
            if timeout.expired:
                continue  # release the coordinator lock

            resource = acquired_resources[resource_path]
            if resource.acquired == "<broken>":
                continue
            place = self.places.get(resource.acquired)
            print(f"should release {resource} for {place}?")

            if place is None:
                logging.warning("resource %s claims to be acquired by unknown place", resource)
            elif not place.acquired:
                logging.warning("resource %s claims to be acquired by unacquired place", resource)
            else:
                continue
            try:
                await self._release_resources(place, [resource])
                del acquired_resources[resource_path]
            except Exception:
                logging.exception("failed to release unused resource %s", resource)
                break

        # find orphaned resources to be acquired
        to_acquire = list(orphaned_resources.keys() - acquired_resources.keys())
        if to_acquire:
            logging.info("synchronize resources: %s orphaned resource(s) should be acquired", len(to_acquire))
            random.shuffle(to_acquire)  # don't get stuck on a problematic resource
        for resource_path in to_acquire:
            if timeout.expired:
                continue  # release the coordinator lock

            resource = orphaned_resources[resource_path]
            if resource.acquired == "<broken>":
                continue
            place = self.places.get(resource.acquired)
            assert place is not None
            assert place.acquired
            print(f"should acquire {resource} for {place}?")

            # is the exporter connected again?
            exporter = self.get_exporter_by_name(resource.path[0])
            if not exporter:
                continue

            # does the resource exist again?
            try:
                new_resource = exporter.groups[resource.path[1]][resource.path[3]]
            except KeyError:
                continue

            if new_resource.acquired:
                # this should only happen when resources become broken
                logging.warning("ignoring acquired/broken resource %s for place %s", new_resource, place.name)
                continue

            try:
                await self._acquire_resource(place, new_resource)
                acquired_resources[new_resource.path] = new_resource
            except Exception:
                logging.exception("failed to reacquire orphaned resource %s for place %s", new_resource, place.name)
                break

        # find orphaned resources to be replaced in the places
        to_replace = set(orphaned_resources.keys() & acquired_resources.keys())
        if to_replace:
            logging.info("synchronize resources: %s orphaned resource(s) should be replaced", len(to_replace))
        for resource_path in set(orphaned_resources.keys() & acquired_resources.keys()):
            oldresource = orphaned_resources[resource_path]
            newresource = acquired_resources[resource_path]
            assert oldresource.acquired == newresource.acquired

            place = self.places.get(newresource.acquired)
            assert place is not None
            assert place.acquired

            idx = place.acquired_resources.index(oldresource)
            place.acquired_resources[idx] = newresource

    @locked
    async def AcquirePlace(self, request, context):
        peer = context.peer()
        name = request.placename
        try:
            username = self.clients[peer].name
        except KeyError:
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Peer {peer} does not have a valid session")
        print(request)

        try:
            place = self.places[name]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {name} does not exist")
        if place.acquired:
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Place {name} is already acquired")
        if place.reservation:
            res = self.reservations[place.reservation]
            if not res.owner == username:
                await context.abort(grpc.StatusCode.PERMISSION_DENIED, f"Place {name} was not reserved for {username}")

        # FIXME use the session object instead? or something else which
        # survives disconnecting clients?
        place.acquired = username
        resources = []
        for _, session in sorted(self.exporters.items()):
            for _, group in sorted(session.groups.items()):
                for _, resource in sorted(group.items()):
                    if not place.hasmatch(resource.path):
                        continue
                    resources.append(resource)
        if not await self._acquire_resources(place, resources):
            # revert earlier change
            place.acquired = None
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Failed to acquire resources for place {name}")
        place.touch()
        self._publish_place(place)
        self.save_later()
        self.schedule_reservations()
        print(f"{place.name}: place acquired by {place.acquired}")
        return labgrid_coordinator_pb2.AcquirePlaceResponse()

    @locked
    async def ReleasePlace(self, request, context):
        name = request.placename
        print(request)
        fromuser = request.fromuser if request.HasField("fromuser") else None
        try:
            place = self.places[name]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {name} does not exist")
        if not place.acquired:
            if fromuser:
                return labgrid_coordinator_pb2.ReleasePlaceResponse()
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Place {name} is not acquired")
        if fromuser and place.acquired != fromuser:
            return labgrid_coordinator_pb2.ReleasePlaceResponse()

        await self._release_resources(place, place.acquired_resources)

        place.acquired = None
        place.allowed = set()
        place.touch()
        self._publish_place(place)
        self.save_later()
        self.schedule_reservations()
        print(f"{place.name}: place released")
        return labgrid_coordinator_pb2.ReleasePlaceResponse()

    @locked
    async def AllowPlace(self, request, context):
        placename = request.placename
        user = request.user
        peer = context.peer()
        try:
            username = self.clients[peer].name
        except KeyError:
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Peer {peer} does not have a valid session")
        try:
            place = self.places[placename]
        except KeyError:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Place {placename} does not exist")
        if not place.acquired:
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Place {placename} is not acquired")
        if not place.acquired == username:
            await context.abort(
                grpc.StatusCode.FAILED_PRECONDITION, f"Place {placename} is not acquired by {username}"
            )
        place.allowed.add(user)
        place.touch()
        self._publish_place(place)
        self.save_later()
        return labgrid_coordinator_pb2.AllowPlaceResponse()

    def _get_places(self):
        return {k: v.asdict() for k, v in self.places.items()}

    @locked
    async def GetPlaces(self, unused_request, unused_context):
        logging.debug("GetPlaces")
        try:
            return labgrid_coordinator_pb2.GetPlacesResponse(places=[x.as_pb2() for x in self.places.values()])
        except Exception:
            logging.exception("error during get places")

    def schedule_reservations(self):
        # The primary information is stored in the reservations and the places
        # only have a copy for convenience.

        # expire reservations
        assert self.lock.locked()
        for res in list(self.reservations.values()):
            if res.state is ReservationState.acquired:
                # acquired reservations do not expire
                res.refresh()
            if not res.expired:
                continue
            if res.state is not ReservationState.expired:
                res.state = ReservationState.expired
                res.allocations.clear()
                res.refresh()
                print(f"reservation ({res.owner}/{res.token}) is now {res.state.name}")
            else:
                del self.reservations[res.token]
                print(f"removed {res.state.name} reservation ({res.owner}/{res.token})")

        # check which places are already allocated and handle state transitions
        allocated_places = set()
        for res in self.reservations.values():
            acquired_places = set()
            for group in list(res.allocations.values()):
                for name in group:
                    place = self.places.get(name)
                    if place is None:
                        # the allocated place was deleted
                        res.state = ReservationState.invalid
                        res.allocations.clear()
                        res.refresh(300)
                        print(f"reservation ({res.owner}/{res.token}) is now {res.state.name}")
                    if place.acquired is not None:
                        acquired_places.add(name)
                    assert name not in allocated_places, "conflicting allocation"
                    allocated_places.add(name)
            if acquired_places and res.state is ReservationState.allocated:
                # an allocated place was acquired
                res.state = ReservationState.acquired
                res.refresh()
                print(f"reservation ({res.owner}/{res.token}) is now {res.state.name}")
            if not acquired_places and res.state is ReservationState.acquired:
                # all allocated places were released
                res.state = ReservationState.allocated
                res.refresh()
                print(f"reservation ({res.owner}/{res.token}) is now {res.state.name}")

        # check which places are available for allocation
        available_places = set()
        for name, place in self.places.items():
            if place.acquired is None and place.reservation is None:
                available_places.add(name)
        assert not (available_places & allocated_places), "inconsistent allocation"
        available_places -= allocated_places

        # check which reservations should be handled, ordered by priority and age
        pending_reservations = []
        for res in sorted(self.reservations.values(), key=lambda x: (-x.prio, x.created)):
            if res.state is not ReservationState.waiting:
                continue
            pending_reservations.append(res)

        # run scheduler
        place_tagsets = []
        for name in available_places:
            tags = set(self.places[name].tags.items())
            # support place names
            tags |= {("name", name)}
            # support place aliases
            place_tagsets.append(TagSet(name, tags))
        filter_tagsets = []
        for res in pending_reservations:
            filter_tagsets.append(TagSet(res.token, set(res.filters["main"].items())))
        allocation = schedule(place_tagsets, filter_tagsets)

        # apply allocations
        for res_token, place_name in allocation.items():
            res = self.reservations[res_token]
            res.allocations = {"main": [place_name]}
            res.state = ReservationState.allocated
            res.refresh()
            print(f"reservation ({res.owner}/{res.token}) is now {res.state.name}")

        # update reservation property of each place and notify
        old_map = {}
        for place in self.places.values():
            old_map[place.name] = place.reservation
            place.reservation = None
        new_map = {}
        for res in self.reservations.values():
            if not res.allocations:
                continue
            assert len(res.allocations) == 1, "only one filter group is implemented"
            for group in res.allocations.values():
                for name in group:
                    assert name not in new_map, "conflicting allocation"
                    new_map[name] = res.token
                    place = self.places.get(name)
                    assert place is not None, "invalid allocation"
                    place.reservation = res.token
        for name in old_map.keys() | new_map.keys():
            if old_map.get(name) != new_map.get(name):
                self._publish_place(self.places[name])

    @locked
    async def CreateReservation(self, request: labgrid_coordinator_pb2.CreateReservationRequest, context):
        peer = context.peer()

        fltrs = {}
        for name, fltr_pb in request.filters.items():
            if name != "main":
                await context.abort(
                    grpc.StatusCode.UNIMPLEMENTED, "Reservations for multiple groups are not implemented yet"
                )
            fltr = fltrs[name] = {}
            for k, v in fltr_pb.filter.items():
                if not TAG_KEY.match(k):
                    await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Key {k} is invalid")
                if not TAG_VAL.match(v):
                    await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Value {v} is invalid")
                fltr[k] = v

        owner = self.clients[peer].name
        res = Reservation(owner=owner, prio=request.prio, filters=fltrs)
        self.reservations[res.token] = res
        self.schedule_reservations()
        return labgrid_coordinator_pb2.CreateReservationResponse(reservation=res.as_pb2())

    @locked
    async def CancelReservation(self, request: labgrid_coordinator_pb2.CancelReservationRequest, context):
        token = request.token
        if not isinstance(token, str) or not token:
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT, f"Invalid token {token}")
        if token not in self.reservations:
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Reservation {token} does not exist")
        del self.reservations[token]
        self.schedule_reservations()
        return labgrid_coordinator_pb2.CancelReservationResponse()

    @locked
    async def PollReservation(self, request: labgrid_coordinator_pb2.PollReservationRequest, context):
        token = request.token
        try:
            res = self.reservations[token]
        except KeyError:
            await context.abort(grpc.StatusCode.FAILED_PRECONDITION, f"Reservation {token} does not exist")
        res.refresh()
        return labgrid_coordinator_pb2.PollReservationResponse(reservation=res.as_pb2())

    @locked
    async def GetReservations(self, request: labgrid_coordinator_pb2.GetReservationsRequest, context):
        reservations = [x.as_pb2() for x in self.reservations.values()]
        return labgrid_coordinator_pb2.GetReservationsResponse(reservations=reservations)


async def serve(listen, cleanup) -> None:
    asyncio.current_task().set_name("coordinator-serve")
    # It seems since https://github.com/grpc/grpc/pull/34647, the
    # ping_timeout_ms default of 60 seconds overrides keepalive_timeout_ms,
    # so set it as well.
    # Use GRPC_VERBOSITY=DEBUG GRPC_TRACE=http_keepalive for debugging.
    channel_options = [
        ("grpc.so_reuseport", 0),  # no load balancing
        ("grpc.keepalive_time_ms", 10000),  # 10 seconds
        ("grpc.keepalive_timeout_ms", 10000),  # 10 seconds
        ("grpc.http2.ping_timeout_ms", 15000),  # 15 seconds
        ("grpc.http2.min_ping_interval_without_data_ms", 5000),
        ("grpc.http2.max_pings_without_data", 0),  # no limit
        ("grpc.keepalive_permit_without_calls", 1),  # allow keepalive pings even when there are no calls
    ]
    server = grpc.aio.server(
        options=channel_options,
    )
    coordinator = Coordinator()
    labgrid_coordinator_pb2_grpc.add_CoordinatorServicer_to_server(coordinator, server)
    # enable reflection for use with grpcurl
    reflection.enable_server_reflection(
        (
            labgrid_coordinator_pb2.DESCRIPTOR.services_by_name["Coordinator"].full_name,
            reflection.SERVICE_NAME,
        ),
        server,
    )
    # optionally enable channelz for use with grpcdebug
    try:
        from grpc_channelz.v1 import channelz

        channelz.add_channelz_servicer(server)
        logging.info("Enabled channelz support")
    except ImportError:
        logging.info("Module grpcio-channelz not available")

    bound = server.add_insecure_port(listen)
    logging.debug("Starting server")
    await server.start()

    if inspect:
        inspect.coordinator = coordinator

    async def server_graceful_shutdown():
        logging.info("Starting graceful shutdown...")
        # Shuts down the server with 0 seconds of grace period. During the
        # grace period, the server won't accept new connections and allow
        # existing RPCs to continue within the grace period.
        await server.stop(5)

    def callback():
        asyncio.ensure_future(server_graceful_shutdown())

    cleanup.append(server_graceful_shutdown())
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, callback)
    loop.add_signal_handler(signal.SIGTERM, callback)
    logging.info("Coordinator ready")
    host, sep, port = listen.rpartition(":")
    if not sep or not port.isdigit():
        host = listen
    print(f"listening on {host}:{bound}", flush=True)
    await server.wait_for_termination()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--listen",
        metavar="HOST:PORT",
        type=str,
        default="[::]:20408",
        help="coordinator listening host and port",
    )
    parser.add_argument("-d", "--debug", action="store_true", default=False, help="enable debug mode")
    parser.add_argument("--pystuck", action="store_true", help="enable pystuck")
    parser.add_argument(
        "--pystuck-port", metavar="PORT", type=int, default=6666, help="use a different pystuck port than 6666"
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    global inspect
    if args.pystuck:
        from types import SimpleNamespace

        inspect = SimpleNamespace()
        inspect.loop = loop

        import pystuck

        pystuck.run_server(port=args.pystuck_port)
    else:
        inspect = None

    cleanup = []
    loop.set_debug(True)
    try:
        loop.run_until_complete(serve(args.listen, cleanup))
    finally:
        if cleanup:
            loop.run_until_complete(*cleanup)


if __name__ == "__main__":
    main()



from labgrid.resource.remote import NetworkUSBTMC
from labgrid.driver.usbtmcdriver import USBTMCDriver


def test_usbtmc_create(target):
    r = NetworkUSBTMC(
        target,
        name=None,
        host="localhost",
        busnum=0,
        devnum=1,
        path="/dev/usbtmc0",
        vendor_id=0x0,
        model_id=0x0,
    )
    d = USBTMCDriver(target, name=None)
    assert isinstance(d, USBTMCDriver)


def test_import_backends():
    import labgrid.driver.usbtmc.keysight_dsox2000



"""
This module implements mounting file systems via communication with udisksd.
"""

import logging
import time

import gi

gi.require_version("UDisks", "2.0")
from gi.repository import GLib, UDisks


class UDisks2Device:
    UNMOUNT_MAX_RETRIES = 5
    UNMOUNT_BUSY_WAIT = 3  # s

    def __init__(self, devpath):
        self._logger = logging.getLogger("Device: ")
        self.devpath = devpath
        self.fs = None

    def _setup(self):
        """Try to find the devpath

        Raises:
            ValueError: no udisks2 device or no filesystem found on devpath
        """
        client = UDisks.Client.new_sync(None)
        manager = client.get_object_manager()
        for obj in manager.get_objects():
            block = obj.get_block()
            if not block:
                continue

            device_path = block.get_cached_property("Device").get_bytestring().decode("utf-8")
            if device_path == self.devpath:
                self.fs = obj.get_filesystem()
                if self.fs is None:
                    raise ValueError(f"no filesystem found on {self.devpath}")

                return

        raise ValueError(f"No udisks2 device found for {self.devpath}")

    def mount(self, readonly=False, retries=0):
        opts = GLib.Variant("a{sv}", {"options": GLib.Variant("s", "ro" if readonly else "rw")})

        try:
            mountpoint = self.fs.call_mount_sync(opts, None)
        except GLib.GError as err:
            if not err.matches(UDisks.error_quark(), UDisks.Error.ALREADY_MOUNTED):
                raise err

            self._logger.warning("Unmounting lazily and remounting %s...", self.devpath)
            self._unmount_lazy()

            mountpoint = self.fs.call_mount_sync(opts, None)

        return mountpoint

    def _unmount_lazy(self):
        opts = GLib.Variant("a{sv}", {"force": GLib.Variant("b", True)})

        try:
            self.fs.call_unmount_sync(opts, None)
        except GLib.GError as err:
            if not err.matches(UDisks.error_quark(), UDisks.Error.NOT_MOUNTED):
                raise err

    def _unmount(self):
        opts = GLib.Variant("a{sv}", {"force": GLib.Variant("b", False)})

        for _ in range(self.UNMOUNT_MAX_RETRIES):
            try:
                self.fs.call_unmount_sync(opts, None)
                return
            except GLib.GError as err:
                if not err.matches(UDisks.error_quark(), UDisks.Error.DEVICE_BUSY):
                    raise err

                self._logger.warning("waiting %s s for busy %s", self.UNMOUNT_BUSY_WAIT, self.devpath)
                time.sleep(self.UNMOUNT_BUSY_WAIT)

        raise TimeoutError("Timeout waiting for device to become non-busy")

    def unmount(self, lazy=False):
        if lazy:
            self._unmount_lazy()
        else:
            self._unmount()


_devs = {}


def _get_udisks2_dev(devpath, retries):
    """Try to get the udisks2 device

    Args:
        devpath (str): Device name
        retries (int): Number of retries to allow

    Raises:
        ValueError: Failed to obtain the device (e.g. does not exist)
    """
    if devpath not in _devs:
        dev = UDisks2Device(devpath=devpath)
        while True:
            try:
                dev._setup()
                break
            except ValueError as exc:
                if "No udisks2 device" not in str(exc) or not retries:
                    raise
                retries -= 1
            dev._logger.warning("udisks2: Retrying %s...", devpath)
            time.sleep(1)

        # Success, so record the new device
        _devs[devpath] = dev
    return _devs[devpath]


def handle_mount(devpath, retries=0):
    dev = _get_udisks2_dev(devpath, retries)
    return dev.mount()


def handle_unmount(devpath, lazy=False):
    dev = _get_udisks2_dev(devpath, 0)
    return dev.unmount(lazy=lazy)


methods = {
    "mount": handle_mount,
    "unmount": handle_unmount,
}



import pytest

pytest.importorskip("onewire")

from labgrid.driver import OneWirePIODriver
from labgrid.resource import OneWirePIO


@pytest.fixture(scope="function")
def onewire_port(target):
    return OneWirePIO(target, "pio", "testhost", "29.123450000000/PIO.6")


@pytest.fixture(scope="function")
def onewire_driver(target, onewire_port, mocker):
    onewire_mock = mocker.patch("onewire.Onewire")
    onewire_mock.return_value.get.return_value = "1"
    s = OneWirePIODriver(target, "pio")
    target.activate(s)
    return s


def test_onewire_resource_instance(target):
    o = OneWirePIO(target, "pio", "testhost", "path")
    assert isinstance(o, OneWirePIO)


def test_onewire_driver_instance(target, onewire_driver):
    isinstance(onewire_driver, OneWirePIODriver)


def test_onewire_set(onewire_driver):
    onewire_driver.set(True)
    onewire_driver._onewire.set.assert_called_once_with("29.123450000000/PIO.6", "1")


def test_onewire_unset(onewire_driver):
    onewire_driver.set(False)
    onewire_driver._onewire.set.assert_called_once_with("29.123450000000/PIO.6", "0")


def test_onewire_get(onewire_driver):
    val = onewire_driver.get()
    onewire_driver._onewire.get.assert_called_once_with("29.123450000000/sensed.6")
    assert val == True



from labgrid.resource.remote import NetworkSiSPMPowerPort
from labgrid.driver.powerdriver import SiSPMPowerDriver


def test_sispm_create(target):
    r = NetworkSiSPMPowerPort(
        target,
        name=None,
        host="localhost",
        busnum=0,
        devnum=1,
        path="0:1",
        vendor_id=0x0,
        model_id=0x0,
        index=1,
    )
    d = SiSPMPowerDriver(target, name=None)
    assert isinstance(d, SiSPMPowerDriver)
